from . import database
from . import utils


class DeckDetails(database.StoredObject):
    """A DeckDetails object represents a deck in a source."""

    def __init__(self, deck_id, source, db_id=None):
        super().__init__("decks", db_id)
        self.deck_id: str = deck_id
        self.source: str = source

    def __hash__(self):
        return hash((self.deck_id, self.source))

    def __repr__(self):
        return (
            f"<DeckDetails deck_id={self.deck_id} source={self.source} "
            f"id={self._id}>"
        )


class Deck(DeckDetails):
    """A Deck object represents a deck downloaded from a source."""

    # The cards held by the deck are (quantity, name) tuples rather than Card
    # objects. They are parsed into Cards before saving.
    def __init__(self, deck_id, source, name, description, **kwargs):
        super().__init__(deck_id, source, db_id=kwargs.get("id"))
        self.name: str = name
        self.description: str = description
        self.main = kwargs.get("main", [])
        self.side = kwargs.get("side", [])
        self.maybe = kwargs.get("maybe", [])
        self.commanders = kwargs.get("commanders", [])

    def __repr__(self):
        return super().__repr__().replace("<DeckDetails", "<Deck")

    def sort_cards(self, cards):
        # Sort alphabetically by card name
        return sorted(cards, key=lambda c: c[1])

    def get_card_names(self, board):
        return [c[1] for c in self.get_board(board)]

    def get_all_card_names(self):
        return set(
            self.get_card_names("main")
            + self.get_card_names("side")
            + self.get_card_names("maybe")
            + self.get_card_names("commanders")
        )

    def get_main_deck(self, include_commanders=False):
        main_deck = self.main
        if include_commanders:
            main_deck += self.commanders

        return self.sort_cards(main_deck)

    def get_sideboard(self, include_commanders=True, include_maybe=True):
        sideboard = self.side[:]
        if include_commanders:
            sideboard += self.commanders
        if include_maybe:
            sideboard += self.maybe
        return self.sort_cards(sideboard)

    def get_board(self, board, default="main"):
        # Note: this is to be used to add cards, not to retrieve them, as it
        # doesn't return the cards in sorted order.

        board = board.strip().lower()
        if board == "commanders":
            return self.commanders
        elif board in ["maybe", "maybeboard"]:
            return self.maybe
        elif board in ["side", "sideboard"]:
            return self.side
        elif board in ["main", "maindeck", "mainboard"]:
            return self.main
        else:
            return self.get_board(default)

    def add_card(self, card, board):
        self.get_board(board).append(card)

    def add_cards(self, cards, board):
        self.get_board(board).extend(cards)


class DeckUpdate:
    """A DeckUpdate represents the last time a Deck was updated on a source."""

    # Because these are not stored anywhere, they don't need a db id.
    def __init__(self, deck, updated):
        self.deck: DeckDetails = deck
        self.updated: int = updated

    def __repr__(self):
        return f"<DeckUpdate deck={repr(self.deck)} updated={self.updated}>"

    def update(self):
        self.updated = utils.time_now()


class Card(database.StoredObject):
    def __init__(
        self,
        name,
        mtgo_id=None,
        is_dfc=False,
        collector_number=None,
        edition=None,
        db_id=None,
    ):
        super().__init__("cards", db_id)
        self.name: str = name
        self.mtgo_id: str = mtgo_id
        self.is_dfc: bool = is_dfc
        self.collector_number: str = collector_number
        self.edition: str = edition

    def __repr__(self):
        return (
            f"<Card name={self.name} mtgo_id={self.mtgo_id} "
            f"is_dfc={self.is_dfc} collector_number={self.collector_number} "
            f"edition={self.edition} id={self._id}>"
        )

    @staticmethod
    def from_record(tup):
        # database record format:
        # (id, name, mtgo_id, is_dfc, collector_number, set, is_reprint)
        _, name, mtgo_id, is_dfc, collector_number, edition, _ = tup
        return Card(
            name, mtgo_id and str(mtgo_id), is_dfc, collector_number, edition
        )



import logging

from . import mode


class Sync(mode.FilterArgsMode):
    def __init__(self):
        super().__init__("s", "sync", "sync decklists", ["profiles"])

    def action(self, cache, args):
        if args.profiles:
            for profile in args.profiles:
                profile.source.ensure_setup(args.interactive, cache)
                profile.download_all()
        else:
            logging.info(
                "No matching profiles to sync."
                ' Add one with "python -m architrice -a"'
            )



import logging

from . import mode


class Latest(mode.FilterArgsMode):
    def __init__(self):
        super().__init__("l", "latest", "download latest deck for user")

    def action(self, cache, args):
        if cache.profiles:
            for profile in cache.profiles:
                profile.source.ensure_setup(args.interactive, cache)
                profile.download_latest()
        else:
            logging.info(
                "No matching profiles to sync."
                ' Add one with "python -m architrice -a".'
            )



import abc
import logging

from .. import database
from .. import deckreprs


class Source(database.KeyStoredObject, abc.ABC):
    # Abstract base class for deck sources, used to have common logging between
    # them.
    #
    # pylint: disable=assignment-from-no-return
    #
    # Private methods are not implemented in this base class, so there are some
    # methods with no returns. This is ok as this class is not directly
    # instantiated.

    def __init__(self, name, short):
        database.KeyStoredObject.__init__(self, short)

        self.name = name
        self.short = short

    def create_deck(self, deck_id, name, description):
        """Create a Deck with relevant information."""
        return deckreprs.Deck(deck_id, self.short, name, description)

    @abc.abstractmethod
    def _get_deck(self, deck_id):
        raise NotImplementedError()

    def get_deck(self, deck_id):
        """Download as `Deck` the deck with id `deck_id` from this source."""
        deck = self._get_deck(deck_id)
        logging.info(f"Downloaded {self.name} deck {deck.name} (id: {deck_id})")
        return deck

    @abc.abstractmethod
    def _get_deck_list(self, username):
        raise NotImplementedError()

    def get_deck_list(self, username):
        """Get a list of `DeckUpdate` for all public decks of `username`."""
        deck_list = self._get_deck_list(username)
        logging.info(
            f"Found {len(deck_list)} decks for {self.name} user {username}."
        )
        return deck_list

    def _get_latest_deck(self, username):
        try:
            return max(self.get_deck_list(username), key=lambda d: d.updated)
        except ValueError:  # max on empty list produces ValueError
            return None

    def get_latest_deck(self, username):
        """Get a `DeckUpdate` for `username`'s most recently updated deck."""
        latest = self._get_latest_deck(username)
        if latest:
            logging.info(
                f"Latest deck for {self.name} user {username} "
                f"has ID {latest.deck.deck_id}."
            )
        else:
            logging.info(
                f"Didn't find any decks for {username} on {self.name}."
            )
        return latest

    @abc.abstractmethod
    def _verify_user(self, username):
        raise NotImplementedError()

    def verify_user(self, username):
        """Verify that user `username` has an account with this source."""
        logging.info(f"Verifying {self.name} user {username}.")
        result = self._verify_user(username)
        if result:
            logging.info("Verification succesful.")
        else:
            logging.error("Verification failed.")
        return result

    def deck_update_from(self, deck_id, time):
        return deckreprs.DeckUpdate(
            deckreprs.DeckDetails(deck_id, self.short), time
        )

    def ensure_setup(self, interactive, cache):
        """Ensure any source-wide setup is complete, throwing on failure."""

        # By default, do nothing. Optionally inheriting sources may perform
        # setup or check that setup is complete.
        pass



#!/bin/python3

import argparse
import logging

from . import modes
from . import sources
from . import targets
from . import utils

APP_NAME = utils.APP_NAME

DESCRIPTION = f"""
{APP_NAME} is a tool to download decks from online sources to local directories.
To set up, run {APP_NAME} with no arguments. This will run a wizard to set up a
link between an online source and a local directory. Future runs of {APP_NAME}
will then download all decklists that have been updated or created since the
last run to that directory. 

To add another download profile beyond this first one, run {APP_NAME} -a.

To delete an existing profile, run {APP_NAME} -d, which will launch a wizard to
do so.

To download only the most recently updated decklist for each profile, run
{APP_NAME} -l.

To set up a new profile or delete a profile without CLI, specify
non-interactivity with the -i or --non-interactive flag and use the flags for
source, user, target, path and name as in 
{APP_NAME} -i -s SOURCE -u USER -t TARGET -p PATH -n NAME -a
Replace -a with -d to delete instead of creating. 

To skip updating decklists while using other functionality, include the -k flag.

To add shortcuts to launch {APP_NAME}, run {APP_NAME} -r.
"""


def add_data_args(parser):
    # These arguments are prefixed with an underscore because they are
    # processed into non-underscored replacements with application objects.
    parser.add_argument(
        "-u", "--user", dest="user", help="set username to download decks of"
    )
    parser.add_argument(
        "-s", "--source", dest="source", help="set source website"
    )
    parser.add_argument(
        "-t", "--target", dest="target", help="set target program"
    )
    parser.add_argument(
        "-p", "--path", dest="path", help="set deck file output directory"
    )
    parser.add_argument(
        "-m",
        "--maybeboard",
        dest="include_maybe",
        help="include maybeboard in output sideboard",
        nargs="?",
        const=1,
    )
    # Any string is admissable for a profile name.
    parser.add_argument("-n", "--name", dest="name", help="set profile name")
    # various flags
    parser.add_argument(
        "-q",
        "--quiet",
        dest="quiet",
        action="store_true",
        help="disable logging to stdout",
    )
    parser.add_argument(
        "-i",
        "--non-interactive",
        dest="interactive",
        action="store_false",
        help="disable interactivity (for scripts)",
    )


def process_args(args):
    args.source = sources.get(args.source)
    args.target = targets.get(args.target)
    args.user = args.user and args.user.strip()
    args.path = utils.expand_path(args.path)
    args.include_maybe = bool(args.include_maybe)

    return args


def main():
    parser = argparse.ArgumentParser(
        description=DESCRIPTION,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    add_data_args(parser)

    for mode in modes.flag_modes:
        parser.add_argument(
            mode.flag,
            mode.long,
            dest=mode.name,
            action="store_true",
            help=mode.help,
        )

    args = process_args(parser.parse_args())
    utils.set_up_logger(int(not args.quiet))

    selected_mode = None
    for mode in modes.flag_modes:
        if getattr(args, mode.name):
            if selected_mode is None:
                selected_mode = mode
            else:
                logging.info(
                    f"Action already selected: {selected_mode.name}."
                    f" Ignoring {mode.flag} flag."
                )

    # Default behaviour: sync decks
    if selected_mode is None:
        selected_mode = modes.Sync()

    selected_mode.main(args)


if __name__ == "__main__":
    main()



from architrice import caching
from . import common
from . import mode


class AddOutput(mode.Mode):
    def __init__(self):
        super().__init__(
            "o",
            "add-output",
            "add an output to a profile",
            ["target", "path", "include_maybe", "profile"],
        )

    def action(self, cache, args):
        common.add_output(
            cache,
            args.interactive,
            args.profile,
            args.target,
            args.path,
            args.include_maybe,
        )

    def load_cache(self, args):
        return caching.Cache.load(
            source=args.source, user=args.user, name=args.name
        )



import re

import bs4
import requests

from .. import utils

from . import source


class TappedOut(source.Source):
    NAME = "Tapped Out"
    SHORT = NAME[0]
    URL_BASE = "https://tappedout.net/"
    MTGA_CARD_REGEX = re.compile(r"^(?P<qty>\d+) (?P<name>.*) \(.*\)( \d+)?$")
    DECK_HREF_TO_ID_REGEX = re.compile(r"^.*/(.*)/$")

    def __init__(self):
        super().__init__(TappedOut.NAME, TappedOut.SHORT)

    def parse_to_cards(self, mtga_string):
        cards = []
        for line in mtga_string.split("\n"):
            m = TappedOut.MTGA_CARD_REGEX.match(line)
            if m:
                cards.append((int(m.group("qty")), m.group("name")))
        return cards

    def deck_to_generic_format(
        self, deck_id, mtga_deck, name, description, commanders
    ):
        d = self.create_deck(deck_id, name, description)

        SIDEBOARD_SEPERATOR = "\n\n"
        if SIDEBOARD_SEPERATOR in mtga_deck:
            main_string, side_string = mtga_deck.split(SIDEBOARD_SEPERATOR)
            d.main.extend(self.parse_to_cards(main_string))
            d.side.extend(self.parse_to_cards(side_string))
        else:
            d.main.extend(self.parse_to_cards(mtga_deck))

        # Move commanders to correct list
        to_move = []
        for tup in d.main:
            _, card = tup
            if card in commanders:
                to_move.append(tup)

        for card in to_move:
            d.main.remove(card)
            d.commanders.append(card)

        return d

    def _get_deck(self, deck_id):
        # TappedOut offers a few export formats, but none of them include deck
        # name, deck description, or specify which cards are commanders.
        # Therefore we scrape the HTML with bs4 instead.

        html = requests.get(
            f"{TappedOut.URL_BASE}mtg-decks/{deck_id}/"
        ).content.decode()

        soup = bs4.BeautifulSoup(html, "html.parser")

        mtga_deck = soup.find(attrs={"id": "mtga-textarea"}).text

        commanders = []
        for tag in soup.select("div.board-col > h3"):
            if "Commander" in tag.text:
                for card in tag.find_next_sibling("ul").select("span > a"):
                    commanders.append(card.get("data-name"))

        PAGE_TITLE_PREFIX = "MTG Deck: "
        name = (
            soup.find("meta", attrs={"property": "og:title"})
            .get("content")
            .replace(PAGE_TITLE_PREFIX, "")
            .strip()
        )

        description = soup.find(
            "meta", attrs={"property": "og:description"}
        ).get("content")

        return self.deck_to_generic_format(
            deck_id, mtga_deck, name, description, commanders
        )

    def age_string_to_timestamp(self, string):
        # No timestamp, so parse the string for an approximation
        now = utils.time_now()
        if string == "Updated a few seconds ago.":
            return now

        m = re.match(
            r"Updated (?P<n>\d+) (?P<unit>minute|hour|day|month|year)s? ago\.",
            string,
        )
        if m:
            return (
                now
                - int(m.group("n"))
                * {
                    "minute": 60,
                    "hour": 60 * 60,
                    "day": 60 * 60 * 24,
                    "month": 60 * 60 * 24 * 28,
                    "year": 60 * 60 * 24 * 365,
                }[m.group("unit")]
            )
        return now

    def get_page_count(self, soup):
        try:
            return int(
                soup.select_one("ul.pagination")
                .find_all("li")[-1]
                .select_one("a.page-btn")
                .text
            )
        except AttributeError:
            # If we hit a None on one of the selects, no pagination ul exists
            # as there is only a single page.
            return 1

    def _get_deck_list(self, username, allpages=True):
        decks = []

        url_base = f"{TappedOut.URL_BASE}users/{username}/mtg-decks/"

        html = requests.get(url_base).content.decode()
        soup = bs4.BeautifulSoup(html, "html.parser")

        pages = 1 if not allpages else self.get_page_count(soup)
        i = 1
        while i <= pages:
            # First page is grabbed outside the loop so that the number of pages
            # can be determined in advance. For other pages we need to download
            # the page now.
            if i > 1:
                html = requests.get(url_base + f"?page={i}").content.decode()
                soup = bs4.BeautifulSoup(html, "html.parser")

            for chunk in utils.group_iterable(soup.select("div.contents"), 3):
                # Each set of three divs is a single deck entry. The first div
                # is the colour breakdown graph, which is not relevant.

                _, name_div, details_div = chunk

                deck_id = re.sub(
                    TappedOut.DECK_HREF_TO_ID_REGEX,
                    r"\1",
                    name_div.select_one("h3.name > a").get("href"),
                )

                for h5 in details_div.select("h5"):
                    if "Updated" in h5.text:
                        updated = self.age_string_to_timestamp(h5.text.strip())
                        break

                decks.append(self.deck_update_from(deck_id, updated))

            i += 1

        return decks

    def _verify_user(self, username):
        return bool(len(self._get_deck_list(username, False)))



from .profile import AddProfile
from .delete import DeleteProfile
from .edit_json import EditJson
from .export_json import ExportJson
from .import_json import ImportJson
from .latest import Latest
from .output import AddOutput
from .relnk import Relnk
from .version import Version
from .clear_cards import ClearCards

from .sync import Sync  # Export

flag_modes = [
    AddProfile(),
    DeleteProfile(),
    EditJson(),
    ExportJson(),
    ImportJson(),
    Latest(),
    AddOutput(),
    Relnk(),
    Version(),
    ClearCards(),
]



from . import mode


class DeleteProfile(mode.FilterArgsMode):
    def __init__(self):
        super().__init__(
            "d",
            "delete",
            "launch wizard or use options to delete a profile",
            ["profile"],
        )

    def action(self, cache, args):
        cache.remove_profile(args.profile)



from . import card_info
from . import target


class XMage(target.Target):
    NAME = "XMage"
    SHORT = NAME[0]
    FILE_EXTENSION = ".dck"
    MAIN_DECK_FORMAT = "{} [{}:{}] {}\n"
    SIDEBOARD_FORMAT = f"SB: {MAIN_DECK_FORMAT}"
    SHORTCUT_NAME = "XMage.lnk"
    EXECUTABLE_NAME = "mage-client"
    SUPPORTS_RELNK = True

    def __init__(self):
        super().__init__(XMage.NAME, XMage.SHORT, XMage.FILE_EXTENSION)

    def format_card_list(self, card_info_map, card_list, sideboard=False):
        format_string = (
            XMage.SIDEBOARD_FORMAT if sideboard else XMage.MAIN_DECK_FORMAT
        )

        card_list_string = ""
        for quantity, name in card_list:
            info = card_info_map.get(name)
            if info is None:  # Skip cards we don't have data for
                continue

            card_list_string += format_string.format(
                quantity,
                info.edition.upper(),
                info.collector_number,
                self.front_face_name(name, card_info_map),
            )

        return card_list_string

    def _save_deck(self, deck, path, include_maybe=False, card_info_map=None):
        # XMage decks have the following format:
        #
        # QTY [SET:COLLECTOR_NUMBER] CARD_NAME
        #   ... for each main deck card
        # SB: QTY [SET:COLLECTOR_NUMBER] CARD_NAME
        #   ... for each sideboard card
        # LAYOUT MAIN:(ROWS, COLS)(NONE,false,50)|([SET:COLLECTOR_NUMBER],)
        #   where each | seperated tuple contains the cards in that cell
        # LAYOUT SIDEBOARD:(ROW< COLS)(NONE,false,50)|([SET:COLLECTOR_NUMBER],)
        #   as with the main deck. These layout specifications are optional and
        #   so Architrice omits them.
        deck_string = self.format_card_list(
            card_info_map, deck.get_main_deck()
        ) + self.format_card_list(
            card_info_map, deck.get_sideboard(include_maybe=include_maybe), True
        )

        with open(path, "w") as f:
            f.write(deck_string)



import functools
import logging
import re

import requests

from .. import database
from .. import deckreprs
from .. import utils

SCRYFALL_BULK_DATA_URL = "https://api.scryfall.com/bulk-data/default-cards"
# Scryfall updates its card list every 24 hours.
# We will update no more frequently than this as it is a large download.
CARD_LIST_UPDATE_INTERVAL = 60 * 60 * 24

# Could consider Ascii normalising card names upon database insertion. This
# would make it easier to find cards with accents in names. However, some
# targets, like Cockatrice, need the accents in card names.
#
# ASCII_NORMALISATION_REGEX = re.compile(r"[^ !&'\(\),\-\./\?A-Z0-9_a-z]")
#
# def ascii_normalise_card_name(name):
#     return re.sub(ASCII_NORMALISATION_REGEX, r"_", name)


def wildcard_vowels(name):
    return re.sub(r"[aeiou]", r"_", name)


def card_info_tuple(card_json):
    dfc_layouts = ["meld", "modal_dfc", "transform"]
    return (
        card_json["name"],
        card_json.get("mtgo_id"),
        card_json["layout"] in dfc_layouts,
        card_json["collector_number"],
        card_json["set"],
        card_json["reprint"],
    )


def save_card_info(data):
    if not isinstance(data, list):
        if data["object"] == "card":
            data = [data]
        elif data["object"] == "list":
            data = data["data"]
        else:
            return

    disallowed_layouts = [
        "art_series",
        "double_faced_token",
        "emblem",
        "planar",
        "scheme",
        "token",
        "vanguard",
    ]

    # Tuple format:
    # (name, mtgo_id, is_dfc, collector_number, edition, reprint)

    records = []

    # Relatively slow way to do it but as it only needs to happen rarely and
    # is paired with a download anyway it isn't really worth optimising.
    for card in data:
        if card["layout"] in disallowed_layouts:
            continue

        records.append(card_info_tuple(card))

    database.insert_many_tuples(
        "cards",
        [
            "name",
            "mtgo_id",
            "is_dfc",
            "collector_number",
            "edition",
            "reprint",
        ],
        records,
        conflict="ignore",
    )
    database.commit()


# Note: this should only be called from one thread at a time.
# Returns bool indicating whether the database was actually updated.
def update_card_list():
    time, url = database.select_one(
        "database_events",
        ["time", "data"],
        id=database.DatabaseEvents.CARD_LIST_UPDATE.value,
    ) or (0, None)
    if utils.time_now() - time < CARD_LIST_UPDATE_INTERVAL:
        return False

    download_info = requests.get(SCRYFALL_BULK_DATA_URL).json()

    database.upsert(
        "database_events",
        id=database.DatabaseEvents.CARD_LIST_UPDATE.value,
        time=utils.time_now(),
        data=download_info["download_uri"],
    )

    if download_info["download_uri"] == url:
        logging.info("Latest Scryfall card list already downloaded.")
        return

    logging.info(
        "Downloading Scryfall card list for card data. Download size: "
        + str(download_info.get("size"))
        + " bytes."
    )

    # ~30MB download, ~230MB uncompressed
    logging.info("This may take a couple of minutes.")
    data = requests.get(download_info["download_uri"]).json()

    save_card_info(data)

    logging.info("Card database update complete.")

    return True


def update_single(name):
    URL_BASE = "https://api.scryfall.com/cards/search"
    resp = requests.get(
        URL_BASE, params={"q": f'!"{name}"', "unique": "prints"}
    )

    if resp.status_code == 200:
        logging.info(f"Downloaded card data for {name}.")
        data = resp.json()
        save_card_info(data)
    elif resp.status_code != 404:
        logging.error(
            f"Failed to query card data for {name}. Status: {resp.status_code}."
        )


@functools.lru_cache(maxsize=None)  # cache to save repeated db queries
def find(name, mtgo_id_required=False, update_if_necessary=True):
    matches = list(database.select("cards", name=name))
    if not matches:
        # Some websites don't include the back face of cards in the name.
        # Luckily, card face names are unique, so we can simply match cards
        # whose name starts with the front face name.
        matches = list(
            database.execute(
                "SELECT * FROM cards WHERE name LIKE ?;", (name + " // %",)
            )
        )

    if not matches:
        # Card names like like Seance are sometimes normalised to ascii and
        # sometimes left with accents. If they're normalised to ascii, we might
        # be able to find them by replacing vowels with wildcards.
        matches = list(
            database.execute(
                "SELECT * FROM cards WHERE name LIKE ?;",
                (wildcard_vowels(name),),
            )
        )

    # Try and get original printing
    for tup in matches:
        _, _, mtgo_id, *_, reprint = tup

        if not reprint and (mtgo_id or not mtgo_id_required):
            return deckreprs.Card.from_record(tup)

    # Settle for any printing
    for tup in matches:
        _, _, mtgo_id, *_ = tup

        if mtgo_id or not mtgo_id_required:
            return deckreprs.Card.from_record(tup)

    if update_if_necessary:
        logging.debug(f"Missing card info for {name}. Updating database.")

        if not update_card_list():
            update_single(name)
        return find(
            name, mtgo_id_required=mtgo_id_required, update_if_necessary=False
        )
    else:
        logging.error(f"Unable to find card info for {name}.")
        return None


def find_many(names, mtgo_id_required=False):
    """Returns a {name: CardInfo} map with all cards in names."""
    database.disable_logging()
    card_info_map = {
        name: find(name, mtgo_id_required=mtgo_id_required) for name in names
    }
    database.enable_logging()
    return card_info_map


def map_from_deck(deck, mtgo_id_required=False):
    """Returns a card info map from a sources.Deck."""
    return find_many(deck.get_all_card_names(), mtgo_id_required)


def map_from_decks(decks, mtgo_id_required=False):
    """Return a card info map with all cards that appear in decks."""
    card_names = set()
    for deck in decks:
        card_names.update(deck.get_all_card_names())

    return find_many(card_names, mtgo_id_required)



import logging
import os
import subprocess
import sys

from .. import utils

from . import cli
from . import mode

APP_NAME = utils.APP_NAME

try:
    # List of common shortcut locations on windows
    # ("Friendly name", "path\\to\\dir")
    SHORTCUT_PATHS = [
        (
            "Start Menu",
            "C:\\ProgramData\\Microsoft\\Windows\\Start Menu\\Programs",
        ),
        (
            "Start Menu",
            os.path.join(os.getenv("USERPROFILE"), "Start Menu", "Programs"),
        ),
        (
            "Task Bar",
            os.path.join(
                os.getenv("APPDATA"),
                "Microsoft",
                "Internet Explorer",
                "Quick Launch",
                "User Pinned",
                "TaskBar",
            ),
        ),
        ("Desktop", os.path.join(os.getenv("USERPROFILE"), "Desktop")),
    ]
except TypeError:
    # not on Windows, getenv returns None, causing join to TypeError
    # we won't use this list on Linux anyway.
    pass

# Snippets used in powershell scripts to read/edit shortcuts
PS_SHORTCUT_SNIPPET = (
    "(New-Object -ComObject WScript.Shell).CreateShortcut('{}')"
)
PS_RELINK_SNIPPET = (
    f"$shortcut = {PS_SHORTCUT_SNIPPET};"
    "$target_path = $shortcut.TargetPath;"
    "$shortcut.TargetPath = '{}';"
    "$shortcut.IconLocation = -join($target_path,',0');"
    "$shortcut.Save();"
)
# Any single quotes in the command must be appropriately escaped,
# use root_format_command instead of directly calling .format
PS_ROOT_SNIPPET = "Start-Process powershell -Verb RunAs -Args '-Command {}'"
PS_COMMAND_SNIPPET = 'powershell -command "{}"'

# Name of the .bat file created to run both apps
BATCH_FILE_NAME = "run_{}.bat"


def batch_file_name(shortcut_name):
    return BATCH_FILE_NAME.format(
        shortcut_name.replace(".lnk", "").lower()
    )  # e.g. Cockatrice.lnk => run_cockatrice.bat.


def create_batch_file(script_name, client_path):
    batch_file_path = os.path.join(utils.DATA_DIR, script_name)

    if not os.path.exists(batch_file_path):
        with open(batch_file_path, "w") as f:
            f.write(
                PS_COMMAND_SNIPPET.format(f"Start-Process '{client_path}'")
                + f"\n{sys.executable} -m architrice\n"
            )

    return batch_file_path


def get_shortcut_target(shortcut_path):
    return (
        subprocess.check_output(
            PS_COMMAND_SNIPPET.format(
                PS_SHORTCUT_SNIPPET.format(shortcut_path) + ".TargetPath"
            )
        )
        .decode()
        .strip()
    )


def root_format_command(command):
    return PS_ROOT_SNIPPET.format(command.replace("'", "''"))


def relink_shortcut(shortcut_path, new_target, as_admin=False):
    command = PS_RELINK_SNIPPET.format(shortcut_path, new_target)
    try:
        subprocess.check_call(
            PS_COMMAND_SNIPPET.format(
                root_format_command(command) if as_admin else command
            ),
            stderr=subprocess.DEVNULL,
        )
    except subprocess.CalledProcessError:
        logging.error(f"Failed to relink shortcut at {shortcut_path}.")
        if not as_admin:
            logging.info("Retrying as admin.")
            relink_shortcut(shortcut_path, new_target, True)
        return
    logging.info(f"Relinked {shortcut_path} to {new_target}.")


def relink_shortcuts(shortcut_name, confirm=False):
    """Relink all shortcuts named `shortcut_name` to also run Architrice."""

    for friendly_name, directory in SHORTCUT_PATHS:
        for sub_directory in os.walk(directory):
            path, _, files = sub_directory
            if not shortcut_name in files:
                continue

            relative_path = os.path.relpath(path, directory)

            if not confirm or cli.get_decision(
                f"Found Cockatrice shortcut on your {friendly_name}"
                + (f" in {relative_path}" if relative_path != "." else "")
                + ". Would you like to update it to run Architrice at launch?"
            ):
                shortcut_path = os.path.join(path, shortcut_name)
                shortcut_target = get_shortcut_target(shortcut_path)
                script_name = batch_file_name(shortcut_name)
                if shortcut_target and not script_name in shortcut_target:
                    script_path = create_batch_file(
                        script_name, shortcut_target
                    )
                    relink_shortcut(shortcut_path, script_path)


class Relnk(mode.Mode):
    def __init__(self):
        super().__init__(
            "r", "relink", "edit shortcuts to run architrice", ["target"]
        )

    def main(args):
        if os.name == "nt":
            target = utils.get_target(args.target, args.interactive)
            if not target:
                logging.info(
                    "Unable to set up shortcuts as no target has been provided."
                )
                return

            if not target.SUPPORTS_RELNK:
                logging.info(
                    "This target doesn't support shortcut configuration."
                )
                return

            relink_shortcuts(
                target.SHORTCUT_NAME,
                not cli.get_decision("Automatically update all shortcuts?"),
            )
        elif os.name == "posix":
            APP_PATH = f"/usr/bin/{APP_NAME}"
            if cli.get_decision(f"Add script to run {APP_NAME} to /usr/bin/?"):
                script_path = os.path.join(utils.DATA_DIR, APP_NAME)
                with open(script_path, "w") as f:
                    f.write(f"{sys.executable} -m {APP_NAME}\n")
                os.chmod(script_path, 0o755)
                subprocess.call(["sudo", "mv", script_path, APP_PATH])
                logging.info(f'Running "{APP_NAME}" will now run {APP_NAME}.')
        else:
            logging.error("Unsupported operating system.")



from . import common
from . import cli
from . import mode


class AddProfile(mode.Mode):
    def __init__(self):
        super().__init__(
            "a",
            "add-profile",
            "launch wizard to add a new profile",
            ["source", "target", "user", "path", "include_maybe", "name"],
        )

    def action(self, cache, args):
        name = args.name
        if (
            name is None
            and args.interactive
            and cli.get_decision("Name this profile?")
        ):
            name = cli.get_string("Profile name")

        profile = cache.build_profile(args.source, args.user, name)

        common.add_output(
            cache,
            args.interactive,
            profile,
            args.target,
            args.path,
            args.include_maybe,
        )

        return profile



import logging

from .. import database

from . import mode


class ClearCards(mode.Mode):
    def __init__(self):
        super().__init__("c", "clear-cards", "clear card info cache", [])

    def action(self, cache, args):
        database.execute("DELETE FROM cards;")
        logging.info("Successfully cleared card data.")



import logging
import json

from . import cli
from . import common_json
from . import mode


def edit_profile_json(cache, profile):
    editing = json.dumps(profile.to_json(), indent=4)

    while True:
        try:
            editing = cli.get_text_editor(editing, "profile.json")
            edited_json = json.loads(editing)
            if common_json.verify_profile_json(edited_json):
                break
        except json.JSONDecodeError:
            logging.error("Failed to parse edited JSON.")

        if not cli.get_decision("Try again?"):
            return

    # In the case that the new profile is redundant with an existing
    # profile, the same object is reused, so we don't want to remove it.
    if common_json.import_profile_json(cache, edited_json) is not profile:
        cache.remove_profile(profile)

    logging.info("Successfully updated profile.")


class EditJson(mode.FilterArgsMode):
    def __init__(self):
        super().__init__(
            "e", "edit-json", "edit a profile as JSON", ["profile"]
        )

    def action(self, cache, args):
        if not args.interactive:
            logging.info(
                "Interactivity required to edit as JSON. Ignoring -e flag."
            )
        else:
            edit_profile_json(cache, args.profile)



import logging
import os

from .. import database
from .. import sources
from .. import targets
from .. import utils

from . import cli


def get_source(name, picker=False):
    if name is not None:
        if not isinstance(name, str):
            return name

        try:
            return sources.get(name, True)
        except ValueError as e:
            logging.error(str(e))
    if picker:
        return source_picker()
    return None


def source_picker():
    return cli.get_choice(
        [s.NAME for s in sources.sourcelist],
        "Download from which supported decklist website?",
        sources.sourcelist,
    )()


def get_target(name, picker=False):
    if name is not None:
        if not isinstance(name, str):
            return name

        try:
            return targets.get(name)
        except ValueError as e:
            logging.error(str(e))
    if picker:
        return target_picker()
    return None


def target_picker():
    return cli.get_choice(
        [t.NAME for t in targets.targetlist],
        "For which supported MtG client?",
        targets.targetlist,
    )()


def get_profile(cache, interactive, prompt="Choose a profile"):
    if not cache.profiles:
        logging.error("No profiles, unable to select one.")
        return None

    if len(cache.profiles) == 1:
        logging.info("Defaulted to only profile which matches criteria.")
        return cache.profiles[0]
    elif interactive:
        return cli.get_choice(
            [str(p) for p in cache.profiles], prompt, cache.profiles
        )
    else:
        logging.error("Multiple profiles match criteria.")
        return None


def get_verified_user(source, user, interactive=False):
    if not user:
        if interactive:
            user = cli.get_string(source.name + " username")
        else:
            return None

    if not (
        database.select_one("users", source=source.short, name=user)
        or source.verify_user(user)
    ):
        if interactive:
            print("Couldn't find any public decks for this user. Try again.")
            return get_verified_user(source, None, True)
        else:
            return None
    return user


def get_output_path(cache, interactive, target, path):
    if path is not None:
        if utils.check_dir(path):
            return path
        else:
            logging.error(
                f"A file exists at {path} so it can't be used as an output "
                "directory."
            )
            if not interactive:
                return None
            path = None

    existing_output_dirs = cache.get_all_output_dirs()
    if existing_output_dirs and cli.get_decision(
        "Use existing output directory?"
    ):
        if len(existing_output_dirs) == 1:
            path = existing_output_dirs[0].path
            logging.info(f"Only one existing directory, defaulting to {path}.")
        else:
            path = cli.get_choice(
                [d.path for d in existing_output_dirs],
                "Which existing directory should be used for these decks?",
            )
    else:
        path = target.suggest_directory()
        if not (
            (os.path.isdir(path))
            and cli.get_decision(
                f"Found {target.name} deck directory at {path}."
                " Output decklists here?"
            )
        ):
            return get_output_path(
                cache, interactive, target, cli.get_path("Output directory")
            )
    return path


def add_output(
    cache, interactive, profile, target=None, path=None, include_maybe=None
):
    if profile is None:
        profile = get_profile(
            cache, interactive, "Add an output to which profile?"
        )
        if not profile:
            logging.error("No profile specified. Unable to add output.")
            return

    target = get_target(target, interactive)
    if not target:
        logging.error("No target specified. Unable to add output.")
        return

    path = get_output_path(cache, interactive, target, path)
    if not path:
        logging.error("No path specified. Unable to add output.")
        return

    if include_maybe is None:
        include_maybe = cli.get_decision(
            "Include maybeboards in the decks downloaded?"
        )

    cache.build_output(profile, target, path, include_maybe)



import asyncio
import concurrent.futures
import logging
import os
import typing

from . import database
from . import deckreprs
from . import sources
from . import targets
from . import utils


class DeckFile(database.StoredObject, deckreprs.DeckUpdate):
    """A DeckFile represents the last time a local Deck file was updated."""

    def __init__(self, deck, updated, file_name, output, db_id=None):
        database.StoredObject.__init__(self, "deck_files", db_id)
        deckreprs.DeckUpdate.__init__(self, deck, updated)
        self.file_name: str = file_name
        self.output: Output = output

    def __repr__(self):
        return (
            super().__repr__().replace("<DeckUpdate", "<DeckFile")[:-1]
            + f" file_name={self.file_name} output={self.output} id={self._id}>"
        )


class OutputDir(database.StoredObject):
    # Keep a dict mapping path to OutputDir so that only one object exists for
    # each output directory.
    output_dirs: typing.Dict[str, "OutputDir"] = {}

    def __init__(self, path, db_id=None):
        super().__init__("output_dirs", db_id)
        self.path: str = path

        # maps hash(Output, DeckDetails) to DeckFile
        self.deck_files: typing.Dict[int, DeckFile] = {}

    def __repr__(self):
        return (
            f"<OutputDir path={self.path} n_deck_files={len(self.deck_files)} "
            f"id={self._id}>"
        )

    def ensure_exists(self):
        if os.path.isfile(self.path):
            raise FileExistsError(
                f"Output directory {self.path} exists and is a file."
            )

        if not os.path.exists(self.path):
            os.makedirs(self.path)

    def key(self, output, deck):
        return hash((output, deck))

    def create_file_name(self, output, deck):
        """Create a file name for a deck. Will be unique in this dir."""

        file_name = suggested_file_name = output.target.create_file_name(
            deck.name
        )

        i = 1
        while any(d.file_name == file_name for d in self.deck_files.values()):
            file_name = suggested_file_name.replace(".", f"_{i}.")
            i += 1

        return file_name

    def add_deck_file(self, output, deck_file):
        self.deck_files[self.key(output, deck_file.deck)] = deck_file

    def get_deck_file(self, output, deck):
        key = self.key(output, deck)
        if key not in self.deck_files:
            self.deck_files[key] = DeckFile(
                deck, 0, self.create_file_name(output, deck), output
            )
        return self.deck_files[key]

    def has_deck_file(self, output, deck):
        return self.key(output, deck) in self.deck_files

    def deck_needs_updating(self, output, deck_update):
        if not deck_update:
            return False

        # Newly tracked deck file
        if not self.has_deck_file(output, deck_update.deck):
            return True

        deck_file = self.get_deck_file(output, deck_update.deck)

        # Deck file was deleted
        if not os.path.exists(os.path.join(self.path, deck_file.file_name)):
            return True

        # Deck has been updated at source
        if deck_update.updated > deck_file.updated:
            return True

        return False

    def store_deck_files(self):
        for deck_file in self.deck_files.values():
            deck_file.store()

    def remove_output(self, output):
        """Delete all DeckFiles associated with an Output."""

        to_delete = []

        for k, v in self.deck_files.items():
            if v.output is output:
                to_delete.append(k)

        for k in to_delete:
            del self.deck_files[k]

    @staticmethod
    def get_all():
        for tup in database.select("output_dirs"):
            db_id, path = tup
            if path not in OutputDir.output_dirs:
                OutputDir.output_dirs[path] = OutputDir(path, db_id)

        return list(OutputDir.output_dirs.values())

    @staticmethod
    def get(path):
        if not path:
            return None

        if path not in OutputDir.output_dirs:
            for output_dir in OutputDir.output_dirs.values():
                if os.path.exists(path) and os.path.samefile(
                    path, output_dir.path
                ):
                    return output_dir

            OutputDir.output_dirs[path] = OutputDir(
                path, database.select_one_column("output_dirs", "id", path=path)
            )
        return OutputDir.output_dirs[path]

    @staticmethod
    def load(db_id):
        if not db_id:
            return None

        path = (database.select_one_column("output_dirs", "path", id=db_id),)
        OutputDir.output_dirs[path] = OutputDir(path, db_id)
        return OutputDir.output_dirs[path]


class Output(database.StoredObject):
    def __init__(
        self, target, output_dir, include_maybe=False, profile=None, db_id=None
    ):
        super().__init__("outputs", db_id)
        self.target: targets.target.Target = target
        self.output_dir: OutputDir = output_dir
        self.include_maybe: bool = include_maybe or False
        self.profile: Profile = profile  # Needed for FK in db

    def __hash__(self):
        # Only needs to be unique to a given output_dir.
        # As an output dir can have at most one output for each target, the
        # target is a perfect hash.
        return hash(self.target.short)

    def __repr__(self):
        return (
            f"<Output target={self.target.short} "
            f"output_dir={repr(self.output_dir)} id={self._id}>"
        )

    def equivalent(self, other):
        return other and (
            other.target is self.target and other.output_dir is self.output_dir
        )

    def set_profile(self, profile):
        self.profile = profile

    def get_updated_deck_file(self, deck):
        deck_file = self.output_dir.get_deck_file(self, deck)
        deck_file.update()

        return deck_file

    def save_deck(self, deck):
        self.output_dir.ensure_exists()

        self.target.save_deck(
            deck,
            os.path.join(
                self.output_dir.path, self.get_updated_deck_file(deck).file_name
            ),
            self.include_maybe,
        )

    def save_decks(self, decks):
        self.output_dir.ensure_exists()

        deck_tuples = []
        for deck in decks:
            deck_file = self.get_updated_deck_file(deck)
            deck_tuples.append(
                (deck, os.path.join(self.output_dir.path, deck_file.file_name))
            )

        self.target.save_decks(deck_tuples, self.include_maybe)

    def deck_needs_updating(self, deck_update):
        return self.output_dir.deck_needs_updating(self, deck_update)

    def decks_to_update(self, deck_updates):
        to_update = []
        for deck_update in deck_updates:
            if self.deck_needs_updating(deck_update):
                to_update.append(deck_update.deck.deck_id)
        return to_update

    def store(self):
        super().store()
        self.output_dir.store_deck_files()

    def delete_stored(self):
        super().delete_stored()
        self.output_dir.remove_output(self)

    def to_json(self):
        return {
            "target": self.target.name,
            "output_dir": self.output_dir.path,
            "include_maybe": self.include_maybe,
        }

    @staticmethod
    def from_json(data):
        return Output(
            targets.get(data["target"], True),
            OutputDir.get(data["output_dir"]),
            data["include_maybe"],
        )


class User(database.StoredObject):
    # Keep a dict mapping (name, source) to User so that only one instance
    # exists for each user.
    users: typing.Dict[typing.Tuple[str, str], "User"] = {}

    def __init__(self, name, source, source_id=None, db_id=None):
        super().__init__("users", db_id)
        self.name: str = name
        self.source: str = source
        self.source_id: str = source_id

    def __hash__(self):
        return hash((self.name, self.source))

    def __str__(self):
        return self.name

    def __repr__(self):
        return (
            f"<User name={self.name} source={self.source} "
            f"source_id={self.source_id}>"
        )

    @staticmethod
    def get_all():
        for tup in database.select("users"):
            db_id, name, short, source_id = tup
            key = (name, short)
            if key not in User.users:
                User.users[key] = User(name, short, source_id, db_id)

        return User.users.values()

    @staticmethod
    def get(name, source: sources.source.Source):
        if not name:
            return None

        key = (name, source.short)
        if key not in User.users:
            User.users[key] = User(
                name,
                source.short,
                None,
                database.select_one_column(
                    "users", "id", name=name, source=source.short
                ),
            )

        return User.users[key]

    @staticmethod
    def load(db_id):
        if not db_id:
            return None

        name, short, source_id = database.select_one(
            "users", ["name", "source", "source_id"], id=db_id
        )
        user = User(name, short, source_id, db_id)
        User.users[(name, short)] = user
        return user


class Profile(database.StoredObject):
    THREAD_POOL_MAX_WORKERS = 12

    def __init__(self, user, name, outputs=None, db_id=None):
        super().__init__("profiles", db_id)
        self.source: sources.source.Source = sources.get(user.source)
        self.user: User = user
        self.name: str = name or None  # Ignore empty string
        self.outputs: typing.List[Output] = []

        if outputs:
            for output in outputs:
                self.add_output(output)

    def __repr__(self):
        return (
            f"<Profile source={self.source.short} user={self.user} "
            f"name={self.name} outputs={self.outputs} id={self._id}>"
        )

    def __str__(self):
        return self.user_string

    @property
    def user_string(self):
        return f"{self.user} on {self.source.name}"

    def equivalent(self, other):
        return (
            other.user == self.user  # same user
            and other.source is self.source  # on same website
            and all(
                any(output.path == o.path for o in self.outputs)
                for output in other.outputs
            )  # with a subset of the outputs
        )

    def add_output(self, output):
        output.set_profile(self)
        if not any(o.equivalent(output) for o in self.outputs):
            self.outputs.append(output)
        else:
            logging.info(
                "Skipping output addition as new output is equivalent to an"
                " existing output."
            )

    def clear_outputs(self):
        for output in self.outputs:
            output.delete_stored()
        self.outputs = []

    def save_deck(self, deck):
        for output in self.outputs:
            output.save_deck(deck)

    def save_decks(self, decks):
        for output in self.outputs:
            output.save_decks(decks)

    def download_deck(self, deck_id):
        logging.debug(f"Downloading {self.source.name} deck {deck_id}.")

        return self.source.get_deck(deck_id)

    # This is asynchronous so that it can use a ThreadPoolExecutor to speed up
    # perfoming many deck requests.
    async def download_decks_pool(self, deck_ids):
        logging.info(
            f"Downloading {len(deck_ids)} decks for {self.user_string}."
        )

        loop = asyncio.get_event_loop()
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=Profile.THREAD_POOL_MAX_WORKERS
        ) as executor:
            futures = [
                loop.run_in_executor(executor, self.download_deck, deck_id)
                for deck_id in deck_ids
            ]
            decks = await asyncio.gather(*futures)

        # Gather all decks and then save synchonously so that we can update the
        # card database first if necessary.
        self.save_decks(decks)

    def download_all(self):
        logging.info(f"Updating all decks for {self.user_string}.")

        decks_to_update = set()
        deck_list = self.source.get_deck_list(self.user.name)
        for output in self.outputs:
            decks_to_update.update(output.decks_to_update(deck_list))

        asyncio.run(self.download_decks_pool(decks_to_update))
        logging.info(f"Successfully updated all decks for {self.user_string}.")

    def download_latest(self):
        latest = self.source.get_latest_deck(self.user.name)
        if any(output.deck_needs_updating(latest) for output in self.outputs):
            logging.info(f"Updating latest deck for {self.user_string}.")
            self.save_deck(self.download_deck(latest.deck.deck_id))
        else:
            logging.info(
                f"{self.source.name} user {self.user}"
                "'s latest deck is up to date."
            )

    def update(self, latest):
        if not self.outputs:
            logging.info(
                f"No outputs match filters for {self.user_string}. Skipping."
            )

        if latest:
            self.download_latest()
        else:
            self.download_all()

    def store(self):
        self.user.store()
        super().store()
        for output in self.outputs:
            output.store()

    def delete_stored(self):
        for output in self.outputs:
            output.delete_stored()
        super().delete_stored()

    def to_json(self):
        return {
            "source": self.source.name,
            "user": self.user.name,
            "name": self.name,
            "outputs": [output.to_json() for output in self.outputs],
        }

    @staticmethod
    def from_json(data):
        return Profile(
            User.get(data["user"], sources.get(data["source"], True)),
            data["name"],
            [Output.from_json(output) for output in data["outputs"]],
        )


class Cache:
    def __init__(self, profiles=None):
        self.profiles: typing.List[Profile] = profiles or []

    def add_profile(self, profile):
        for p in self.profiles:
            if p.equivalent(profile):
                logging.info(
                    f"A profile with identical details already exists, "
                    "skipping creation."
                )
                return p

        self.profiles.append(profile)
        return profile

    def remove_profile(self, profile):
        self.profiles.remove(profile)
        profile.delete_stored()
        logging.info(f"Deleted profile for {profile.user_string}.")

    def build_profile(self, source, user, name=None):
        return self.add_profile(Profile(User.get(user, source), name, []))

    def build_output(self, profile, target, path, include_maybe):
        profile.add_output(Output(target, OutputDir.get(path), include_maybe))

    def get_all_output_dirs(self):
        return OutputDir.get_all()

    def save(self):
        logging.debug("Saving cache to database.")

        if not utils.DEBUG:
            database.disable_logging()

        for profile in self.profiles:
            profile.store()

        database.enable_logging()
        database.commit()
        logging.debug("Successfullly saved cache, closing connection.")
        database.close()

    def load_string_value(self, key):
        tup = database.select_one("string_values", key=key)
        if tup:
            return tup[1]  # (key, value)
        else:
            return None

    def save_string_value(self, key, value):
        database.upsert("string_values", key=key, value=value)

    @staticmethod
    def load(
        source=None,
        target=None,
        user=None,
        path=None,
        name=None,
    ):
        """Load all relevant data into memory from the database."""
        database.init()

        output_dirs = []
        for tup in database.select_ignore_none("output_dirs", path=path):
            db_id, path = tup
            output_dirs.append(OutputDir(path, db_id))

        profiles = []

        query = (
            "SELECT p.id, p.user, p.name "
            "FROM profiles p LEFT JOIN users u ON p.user = u.id"
        )
        arguments = []
        conditions = []
        if user:
            conditions.append("UPPER(u.name) = UPPER(?)")
            arguments.append(user)
        if source:
            conditions.append("u.source = ?")
            arguments.append(source.short)
        if name:
            conditions.append("p.name = ?")
            arguments.append(name)
        if conditions:
            query += " WHERE " + " AND ".join(conditions)
        for tup in database.execute(query + ";", arguments):
            profile_db_id, profile_user, profile_name = tup

            outputs = []
            for tup in database.select_ignore_none(
                "outputs",
                target=getattr(target, "short", None),
                profile=profile_db_id,
            ):
                (
                    output_db_id,
                    output_target,
                    output_dir_id,
                    _,
                    output_include_maybe,
                ) = tup

                for output_dir in output_dirs:
                    if output_dir.id == output_dir_id:
                        break
                else:
                    raise LookupError(
                        f"Failed to find dir with id {output_dir_id}."
                    )

                output = Output(
                    targets.get(output_target),
                    output_dir,
                    bool(output_include_maybe),
                    db_id=output_db_id,
                )
                outputs.append(output)

                for tup in database.execute(
                    "SELECT "
                    "d.id, d.deck_id, d.source, df.id, df.file_name, df.updated"
                    " FROM deck_files df LEFT JOIN decks d ON df.deck = d.id "
                    "WHERE df.output = ?;",
                    (output.id,),
                ):
                    (
                        d_db_id,
                        d_deck_id,
                        d_source,
                        df_db_id,
                        df_file_name,
                        df_updated,
                    ) = tup
                    output.output_dir.add_deck_file(
                        output,
                        DeckFile(
                            deckreprs.DeckDetails(d_deck_id, d_source, d_db_id),
                            df_updated,
                            df_file_name,
                            output,
                            df_db_id,
                        ),
                    )

            profiles.append(
                Profile(
                    User.load(profile_user),
                    profile_name,
                    outputs,
                    profile_db_id,
                )
            )

        return Cache(profiles)



import logging

from .. import caching

from . import common


class Mode:
    ARGUMENT_RESOLUTION_ORDER = [
        "source",
        "target",
        "user",
        "path",
        "include_maybe",
        "name",
        "profile",
        "profiles",
    ]

    def __init__(self, flag, name, explanation, required_args=None):
        self.short = flag
        self.name = name
        self.help = explanation

        # Sort args so that missing args are resolved in the right order.
        self.required_args = sorted(
            required_args or [],
            key=lambda arg: Mode.ARGUMENT_RESOLUTION_ORDER.index(arg),
        )

    @property
    def flag(self):
        return "-" + self.short

    @property
    def long(self):
        return "--" + self.name

    def has_all_args(self, args):
        for arg in self.required_args:
            if getattr(args, arg) is None:
                return False
        return True

    def resolve_missing_arg(self, cache, arg, args):
        if arg == "source":
            return common.get_source(args.source, args.interactive)
        elif arg == "target":
            return common.get_target(args.target, args.interactive)
        elif arg == "user":
            args.source.ensure_setup(args.interactive, cache)
            return common.get_verified_user(
                args.source, args.user, args.interactive
            )
        elif arg == "path":
            return common.get_output_path(
                cache, args.interactive, args.target, args.path
            )
        elif arg == "include_maybe":
            return True  # Defaults to True
        elif arg == "name":
            return None  # Always optional
        elif arg == "profile":
            return common.get_profile(cache, args.interactive)
        elif arg == "profiles":
            return cache.profiles
        else:
            raise ValueError(f"Can't resolve missing argument: {arg}.")

    def arg_is_optional(self, arg):
        return arg == "name"

    def ensure_all_args(self, cache, args):
        for arg in self.required_args:
            if getattr(args, arg, None) is None:
                value = self.resolve_missing_arg(cache, arg, args)
                if value is None and not self.arg_is_optional(arg):
                    logging.error(
                        f"Missing argument: {arg}."
                        f' Unable to complete action "{self.name}".'
                    )
                    return False
                setattr(args, arg, value)
        return True

    def action(self, cache, args):
        raise NotImplementedError()

    def load_cache(self, args):
        return caching.Cache.load()

    def main(self, args):
        cache = self.load_cache(args)
        if self.ensure_all_args(cache, args):
            self.action(cache, args)
        cache.save()


class FilterArgsMode(Mode):
    def load_cache(self, args):
        return caching.Cache.load(
            args.source,
            args.target,
            args.user,
            args.path,
            args.name,
        )



import logging
import os
import subprocess
import tempfile

from .. import utils

PROMPT = "> "

# Text editing on linux is done through tempfile
EDITOR_ENVIRONMENT_VARIABLE = "EDITOR"
EDITOR_FALLBACK = "nano" if os.name == "posix" else "notepad"


def get_choice(options, prompt, values=None):
    print(prompt)
    for i, option in enumerate(options):
        print(f"\t[{i + 1}] {option}")

    FAILURE_MESSAGE = "Please enter the number of one of the options."

    while True:
        choice = input(PROMPT).strip()

        if choice in options:
            i = options.index(choice)
        elif choice.isnumeric() and 0 < int(choice) <= len(options):
            i = int(choice) - 1
        else:
            print(FAILURE_MESSAGE)
            continue

        if values:
            return values[i]
        return options[i]


def get_decision(prompt, default=True):
    opts = (
        "(" + ("Y" if default else "y") + "/" + ("n" if default else "N") + ")"
    )

    d = input(f"{prompt} {opts} {PROMPT}").strip().lower()
    if d in [
        "y",
        "yes",
    ]:
        return True
    elif default and not d:
        return True
    else:
        return False


def get_string(prompt):
    while True:
        string = input(f"{prompt} {PROMPT}")
        if string:
            return string.strip()


def get_path(prompt):
    return utils.expand_path(input(f"{prompt} {PROMPT}"))


def get_text_editor_posix(default, editor, file_name):
    with tempfile.NamedTemporaryFile(suffix=file_name) as tf:
        tf.write(default.encode())
        tf.flush()
        subprocess.call(
            [
                editor,
                tf.name,
            ]
        )
        tf.seek(0)
        return tf.read().decode()


def get_text_editor_nt(default, editor, file_name):
    path = os.path.join(utils.DATA_DIR, file_name)
    with open(path, "w") as f:
        f.write(default)

    subprocess.call([editor, path], shell=True)

    with open(path, "r") as f:
        return f.read()
    os.remove(path)


def get_text_editor(default="", file_name=None):
    editor = os.environ.get(EDITOR_ENVIRONMENT_VARIABLE, EDITOR_FALLBACK)

    if os.name == "posix":
        return get_text_editor_posix(default, editor, file_name)
    elif os.name == "nt":
        return get_text_editor_nt(default, editor, file_name)
    else:
        logging.error("Unsupported operating system.")



import json

from . import mode


class ExportJson(mode.FilterArgsMode):
    def __init__(self):
        super().__init__(
            "j", "export-json", "export matching profiles as JSON", ["profiles"]
        )

    def action(self, cache, args):
        print(
            json.dumps(
                [profile.to_json() for profile in cache.profiles], indent=4
            )
        )



import logging

from .. import sources
from .. import targets
from .. import utils

from . import common


def verify_output_json(output, i="\b"):
    if not "target" in output:
        logging.error(f"Output {i} is missing a target.")
        return False
    elif not isinstance(output["target"], str):
        logging.error("Output targets must be strings.")
        return False

    try:
        targets.get(output["target"], True)
    except ValueError as e:
        logging.error(str(e))
        return False

    if not "output_dir" in output:
        logging.error(f"Output {i} is missing an output directory.")
        return False
    elif not isinstance(output["output_dir"], str):
        logging.error("Output directories must be strings.")
        return False

    output["output_dir"] = utils.expand_path(output["output_dir"])
    if not utils.check_dir(output["output_dir"]):
        logging.error(f"Output directory {i} already exists and is a file.")
        return False

    if "include_maybe" in output:
        if not isinstance(output["include_maybe"], bool):
            logging.error(
                "The include_maybe flag of an Output must be a string."
            )
            return False
    else:
        output["include_maybe"] = False

    return True


def verify_profile_json(data):
    if not "source" in data:
        logging.error("Profile is missing a source.")
        return False
    elif not isinstance(data["source"], str):
        logging.error("Source must be a string.")
        return False

    try:
        source = sources.get(data["source"], True)
    except ValueError as e:
        logging.error(str(e))
        return False

    if not "user" in data:
        logging.error("Profile is missing a user.")
        return False
    elif not isinstance(data["user"], str):
        logging.error("User must be a string.")
        return False

    if common.get_verified_user(source, data.get("user")) is None:
        return False

    if "name" in data and not (
        data["name"] is None or isinstance(data["name"], str)
    ):
        logging.error("Name must be a string.")
        return False

    if not "outputs" in data:
        data["outputs"] = []

    if not isinstance(data["outputs"], list):
        logging.error("Outputs must be in a list.")
        return False

    for i, output in enumerate(data.get("outputs")):
        if not verify_output_json(output, i):
            return False

    return True


def import_profile_json(cache, profile_json):
    profile = cache.build_profile(
        sources.get(profile_json["source"]),
        profile_json["user"],
        profile_json["name"],
    )

    # In the event profile is an existing profile, we want to clear it's outputs
    # as this process overwrites them. If it's a new profile it won't have any
    # outputs and this is a no-op.
    profile.clear_outputs()

    for output in profile_json["outputs"]:
        cache.build_output(
            profile,
            targets.get(output["target"]),
            output["output_dir"],
            output["include_maybe"],
        )

    return profile



import architrice

import http.server
import os
import threading
import urllib

SILENT = True

DIRECTORY = os.path.join(os.path.dirname(__file__), "web")


class RequestHandler(http.server.SimpleHTTPRequestHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, directory=DIRECTORY, **kwargs)

    def send_head(self):
        # This is a modified version of the source from
        # https://github.com/python/cpython/blob/main/Lib/http/server.py#L670
        # which allows files named like URL parameters to mimic an API
        #
        # The naming convention for such files is like so:
        # ?owner=Test&ownerexact=true
        # would become
        # owner=Test_ownerexact=true

        path = self.translate_path(self.path)

        f = None

        parts = urllib.parse.urlsplit(self.path)

        query_file = os.path.join(path, parts.query.replace("&", "_"))

        if os.path.isfile(query_file):
            path = query_file
        elif path.endswith("/") and os.path.isfile(path[:-1]):
            # allow trailing /
            path = path[:-1]
        elif os.path.isdir(path):
            if not parts.path.endswith("/"):
                self.send_response(http.server.HTTPStatus.MOVED_PERMANENTLY)
                new_parts = (
                    parts[0],
                    parts[1],
                    parts[2] + "/",
                    parts[3],
                    parts[4],
                )
                new_url = urllib.parse.urlunsplit(new_parts)
                self.send_header("Location", new_url)
                self.send_header("Content-Length", "0")
                self.end_headers()
                return None

            for index in "index.html", "index.htm":
                index = os.path.join(path, index)
                if os.path.exists(index):
                    path = index
                    break
            else:
                return self.list_directory(path)

        try:
            f = open(path, "rb")
        except OSError:
            self.send_error(http.server.HTTPStatus.NOT_FOUND, "File not found")
            return None

        ctype = self.guess_type(path)
        try:
            fs = os.fstat(f.fileno())
            self.send_response(http.server.HTTPStatus.OK)
            self.send_header("Content-type", ctype)
            self.send_header("Content-Length", str(fs[6]))
            self.send_header(
                "Last-Modified", self.date_time_string(fs.st_mtime)
            )
            self.end_headers()
            return f
        except:
            f.close()
            raise

    def log_request(self, *args, **kwargs):
        if not SILENT:
            super().log_request(*args, **kwargs)


PORT = 8192

server = http.server.HTTPServer(("", PORT), RequestHandler)

addr, port = server.server_address
URL = f"http://{addr}:{port}/"

thread = threading.Thread(target=server.serve_forever)


def run():
    thread.start()


def stop():
    server.shutdown()
    server.server_close()
    thread.join()

    for source in architrice.sources.sourcelist:
        source.URL_BASE = source._URL_BASE

    architrice.targets.card_info.SCRYFALL_BULK_DATA_URL = (
        architrice.targets.card_info._SCRYFALL_BULK_DATA_URL
    )


def mock():
    # Remap sources
    for source in architrice.sources.sourcelist:
        source._URL_BASE = source.URL_BASE
        source.URL_BASE = URL

    # Remap card_info
    architrice.targets.card_info._SCRYFALL_BULK_DATA_URL = (
        architrice.targets.card_info.SCRYFALL_BULK_DATA_URL
    )
    architrice.targets.card_info.SCRYFALL_BULK_DATA_URL = (
        URL + "bulk-data/default-cards"
    )

    run()



import requests

from .. import utils

from . import source


class Archidekt(source.Source):
    NAME = "Archidekt"
    SHORT = NAME[0]
    URL_BASE = "https://archidekt.com/"

    def __init__(self):
        super().__init__(Archidekt.NAME, Archidekt.SHORT)

    def format_api_request(self, path):
        return Archidekt.URL_BASE + "api/decks/" + path

    def deck_to_generic_format(self, deck_id, deck):
        d = self.create_deck(deck_id, deck["name"], deck["description"])

        for card in deck["cards"]:
            c = (card["quantity"], card["card"]["oracleCard"]["name"])

            if "Commander" in card["categories"]:
                d.commanders.append(c)
            elif "Maybeboard" in card["categories"]:
                d.maybe.append(c)
            elif "Sideboard" in card["categories"]:
                d.side.append(c)
            else:
                d.main.append(c)

        return d

    def _get_deck(self, deck_id, small=True):
        return self.deck_to_generic_format(
            deck_id,
            requests.get(
                self.format_api_request(
                    deck_id + "/" + "small/" if small else "/"
                ),
                params={"format": "json"},
            ).json(),
        )

    def deck_list_to_generic_format(self, decks):
        ret = []
        for deck in decks:
            ret.append(
                self.deck_update_from(
                    str(deck["id"]), utils.parse_iso_8601(deck["updatedAt"])
                )
            )
        return ret

    def _get_deck_list(self, username, allpages=True):
        decks = []
        url = self.format_api_request(
            f"cards/?owner={username}&ownerexact=true"
        )
        while url:
            j = requests.get(url).json()
            decks.extend(j["results"])

            if not allpages:
                break

            url = j["next"]

        return self.deck_list_to_generic_format(decks)

    def _get_latest_deck(self, username):
        try:
            # By passing allpages=False, we avoid grabbing unnecessary pages of
            # results, as Archidekt returns results presorted by updated time.
            return max(
                self._get_deck_list(username, False), key=lambda d: d.updated
            )
        except ValueError:
            return None

    def _verify_user(self, username):
        return bool(len(self._get_deck_list(username, False)))



import architrice

from . import mode


class Version(mode.Mode):
    def __init__(self):
        super().__init__("v", "version", "print version and exit")

    def main(self, args):
        print(architrice.__version__)



import os

from .cockatrice import Cockatrice
from .generic import Generic
from .mtgo import Mtgo
from .xmage import XMage

targetlist = [Cockatrice, Generic, Mtgo, XMage]

# singleton cache; sources have no data so only a single instance is required
_targets = {}


def get(name, error_on_fail=False):
    if name:
        name = name.lower()  # case insensitivity
        for target in targetlist:
            if target.NAME.lower() == name or target.SHORT.lower() == name:
                if target.SHORT not in _targets:
                    _targets[target.SHORT] = target()
                return _targets[target.SHORT]

    if error_on_fail:
        raise ValueError(f'No target matching "{name}" exists.')
    return None


def get_all():
    return list(map(lambda t: get(t.NAME), targetlist))



import logging
import os
import xml.etree.cElementTree as et

from .. import utils

from . import target


class Mtgo(target.Target):
    NAME = "MTGO"
    SHORT = "M"
    DECK_DIRECTORYS = (
        [
            utils.expand_path(
                os.path.join(
                    os.getenv("APPDATA"),
                    "Wizards of the Coast",
                    "Magic Online",
                    "3.0",
                    "Decks",
                )
            ),
            utils.expand_path(
                os.path.join(
                    "C:",
                    "Program Files",
                    "Wizards of the Coast",
                    "Magic Online",
                    "Decks",
                )
            ),
        ]
        if os.name == "nt"
        else []
    )
    DECK_FILE_EXTENSION = ".dek"
    SHORTCUT_NAME = "Magic The Gathering Online.lnk"
    SUPPORTS_RELNK = True

    def __init__(self):
        super().__init__(Mtgo.NAME, Mtgo.SHORT, Mtgo.DECK_FILE_EXTENSION)
        self.mtgo_id_required = True

    def suggest_directory(self):
        for directory in Mtgo.DECK_DIRECTORYS:
            if os.path.exists(directory):
                return directory
        return super().suggest_directory()

    def _save_deck(self, deck, path, include_maybe, card_info_map=None):
        return deck_to_xml(deck, path, include_maybe, card_info_map)


def mtgo_name(name):
    return name.partition("//")[0].strip()


def add_card(root, quantity, name, card_info_map, in_sideboard=False):
    info = card_info_map.get(name)
    if info and info.mtgo_id:
        et.SubElement(
            root,
            "Cards",
            {
                "CatID": info.mtgo_id,
                "Quantity": str(quantity),
                "Sideboard": "true" if in_sideboard else "false",
                "Name": mtgo_name(name),
                "Annotation": "0",
            },
        )
    else:
        logging.info(
            f"Couldn't find MTGO data for {name}." " It may not exist on MTGO."
        )


def deck_to_xml(deck, outfile, include_maybe, card_info_map):
    root = et.Element(
        "Deck",
        {
            "xmlns:xsd": "http://www.w3.org/2001/XMLSchema",
            "xmlns:xsi": "http://www.w3.org/2001/XMLSchema-instance",
        },
    )  # xmlns declaration that MTGO writes in its .dek files.

    et.SubElement(root, "NetDeckID").text = "0"
    et.SubElement(root, "PreconstructedDeckID").text = "0"

    for quantity, name in deck.get_main_deck():
        add_card(root, quantity, name, card_info_map)
    for quantity, name in deck.get_sideboard(include_maybe=include_maybe):
        add_card(root, quantity, name, card_info_map, True)

    et.ElementTree(root).write(outfile, xml_declaration=True, encoding="utf-8")



import dataclasses
import enum
import logging
import os
import sqlite3
import traceback
import typing

from . import utils

DATABASE_FILE = "architrice.db"


class Database:
    USER_VERSION = 2

    def __init__(self, tables=None):
        self.conn = None
        self.file: str = None
        self.tables_to_init = tables
        self.tables = {}
        self.log = True

    def init(self, database_file, initial_setup=False):
        """Connect to and set up the database for user."""
        self.file = database_file
        self.conn = sqlite3.connect(self.file)

        logging.debug("Connected to database.")

        if initial_setup:
            self.execute(f"PRAGMA user_version = {Database.USER_VERSION};")
        self.execute("PRAGMA foreign_keys = ON;")

        if self.tables_to_init is not None:
            for table in self.tables_to_init:
                self.add_table(table, initial_setup)

        self.tables_to_init = None

        # execute returns a cursor, which becomes a list of tuples,
        # the first of which is (version,).
        version = list(self.execute("PRAGMA user_version;"))[0][0]
        
        self.migrate(version)

    def migrate(self, version):
        """Update the database schema if necessary."""

        if version == 0:
            logging.debug("Migrating database from version 0 to version 1.")
            self.execute("ALTER TABLE outputs ADD COLUMN include_maybe INTEGER")
            self.execute("PRAGMA user_version = 1;")
            version = 1
        if version == 1:
            logging.debug("Migrating database from version 1 to version 2.")
            self.add_table(self.tables["string_values"], True)
            self.execute("PRAGMA user_version = 2;")
            version = 2

    def add_table(self, table, create=False):
        """Add a Table to the database, creating it if necessary."""
        table.set_db(self, create)
        self.tables[table.name] = table

    def insert(self, table, **kwargs):
        """Execute an INSERT into table using kwarg keys and values."""
        return self.tables[table].insert(**kwargs)

    def upsert(self, table, **kwargs):
        """Execute an INSERT into table, updating on conflict."""
        kwargs["conflict"] = "update"
        return self.insert(table, **kwargs)

    def insert_many(self, table, **kwargs):
        """Execute many INSERTs into table, using kwarg keys and value lists."""
        self.tables[table].insert_many(**kwargs)

    def insert_many_tuples(self, table, columns, tuples, conflict=None):
        """Execute many INSERT INTO INTO table (columns) VALUES(tuples)"""
        self.execute_many(
            self.tables[table].insert_command(columns, columns, conflict),
            tuples,
        )

    def upsert_many(self, table, **kwargs):
        """Execute many upserts into table."""
        kwargs["conflict"] = "update"
        self.tables[table].insert_many(**kwargs)

    def select(self, table, columns="*", **kwargs):
        """SELECT columns FROM table WHERE kwargs keys = kwarg values."""
        if isinstance(columns, list):
            columns = ", ".join(columns)

        # Note: returns cursor; use list(db.select(...)) to get raw values.
        return self.tables[table].select(columns, **kwargs)

    def select_one(self, table, columns="*", **kwargs):
        """Return the first tuple resulting from this select."""
        result = list(self.select(table, columns, **kwargs))
        if result:
            return result[0]
        return None

    def select_one_column(self, table, column, **kwargs):
        """Return the first column of the first tuple from this select."""
        result = self.select_one(table, column, **kwargs)
        if result:
            return result[0]
        return None

    def select_ignore_none(self, table, columns="*", **kwargs):
        """Perform a SELECT, ignoring all kwargs which are None."""
        none_values = []
        for key in kwargs:
            if kwargs[key] is None:
                none_values.append(key)

        for key in none_values:
            del kwargs[key]

        return self.select(table, columns, **kwargs)

    def select_where_in(self, table, field, values, columns="*"):
        """SELECT columns FROM table WHERE field in values"""

    def delete(self, table, **kwargs):
        """DELETE FROM table WHERE kwarg keys = kwarg values"""
        self.tables[table].delete(**kwargs)

    def update(self, table, updates, where):
        """UPDATE table SET updates WHERE where"""
        self.tables[table].update(updates, where)

    def execute_ret(self, is_insert, cursor, result):
        if is_insert:
            if cursor.rowcount:
                return cursor.lastrowid
            else:
                return None
        else:
            return result

    def execute(self, command, tup=None):
        """Execute an SQL command, logging the command and data."""

        cursor = self.conn.cursor()
        is_insert = command.startswith("INSERT")

        if tup:
            try:
                if self.log:
                    logging.debug(
                        f"Executing database command: {command} with values "
                        f"{tup}."
                    )

                return self.execute_ret(
                    is_insert, cursor, cursor.execute(command, tup)
                )
            except sqlite3.Error as e:
                logging.debug(
                    f"Database errored on command {command} with values {tup}"
                )
                logging.error(f"Database error: {str(e)}")
                if utils.DEBUG:
                    traceback.print_stack()
                exit()

        if self.log:
            logging.debug(f"Executing database command: {command}")
        try:
            return self.execute_ret(is_insert, cursor, cursor.execute(command))
        except sqlite3.Error as e:
            logging.debug(f"Database errored on command {command}")
            logging.error(f"Database error: {str(e)}.")
            if utils.DEBUG:
                traceback.print_stack()
            exit()

    def execute_many(self, command, tups):
        """Execute many SQL commands."""
        if self.log:
            logging.debug(f"Executing many with command: {command}")
        return self.conn.executemany(command, tups)

    def commit(self):
        """Commit database changes."""
        self.conn.commit()

    def close(self):
        """Close the database connection."""
        self.conn.close()

    def enable_logging(self):
        """Enable command logging."""
        self.log = True

    def disable_logging(self):
        """Disable command logging."""
        self.log = False


@dataclasses.dataclass
class Column:
    name: str  # note: conflict and columns are reserved names
    datatype: str
    primary_key: bool = False
    references: str = None  # Foreign key to this table. Cascade delete.
    not_null: bool = False
    unique: bool = False  # Is this column unique
    index_on: bool = False  # Create an index on this column?

    def __str__(self):
        column_def = f"{self.name} {self.datatype}"
        if self.not_null:
            column_def += " NOT NULL"
        if self.primary_key:
            column_def += " PRIMARY KEY"
        if self.unique:
            column_def += " UNIQUE"
        if self.references:
            column_def += f" REFERENCES {self.references} ON DELETE CASCADE"
        return column_def

    @staticmethod
    def foreign_key(name, datatype):
        return Column(name, datatype, references=name + "s", not_null=True)

class Table:
    def __init__(self, name, columns, constraints=None, db=None):
        self.name: str = name
        self.columns: typing.List[Column] = columns
        self.constraints: typing.List[str] = constraints or []
        self.set_db(db)

    def set_db(self, db, create=True):
        self.db = db
        if db and create:
            self.create()

    def primary_key(self):
        for c in self.columns:
            if c.primary_key:
                return c

    def create(self):
        self.db.execute(
            f"CREATE TABLE IF NOT EXISTS {self.name} ("
            + ", ".join(str(c) for c in self.columns + self.constraints)
            + ");"
        )

        for c in self.columns:
            if c.index_on:
                self.db.execute(
                    f"CREATE INDEX IF NOT EXISTS idx_{self.name}_{c.name} "
                    f"ON {self.name} ({c.name});"
                )

    def column_string(self, columns):
        return "(" + ", ".join(columns) + ")"

    def substitution_string(self, iterable):
        return "(" + ("?, " * len(iterable))[:-2] + ")"

    def on_conflict_update_string(self, conflict_columns, update_columns):
        if len(update_columns) == 0:
            update_string = " DO NOTHING"
        else:
            update_string = (
                " DO UPDATE SET "
                + self.column_string(update_columns)
                + " = "
                + self.substitution_string(update_columns)
            )

        return (
            " ON CONFLICT"
            + self.column_string(conflict_columns)
            + update_string
        )

    def update_columns(self, conflict_columns, column_names, arguments):
        # Note: this mutates the argument list to add additional arguments
        # for the new updates.

        update_columns = [
            c for c in column_names if c not in conflict_columns and c != "id"
        ]
        for c in update_columns:
            arguments.append(arguments[column_names.index(c)])
        return update_columns

    def create_upsert_string(self, column_names, arguments):
        # Note:
        # sqlite only supports a single on conflict term, so this prioritises
        # first multi-column unique statements, then unique columns and finally
        # primary keys.

        for c in self.constraints:
            if "UNIQUE" in c:
                conflict_columns = (
                    c.replace("UNIQUE(", "").replace(")", "").split(", ")
                )

                # Only handle conflicts in columns which could actually occur
                if any(c in column_names for c in conflict_columns):
                    return self.on_conflict_update_string(
                        conflict_columns,
                        self.update_columns(
                            conflict_columns, column_names, arguments
                        ),
                    )

        for c in self.columns:
            if c.unique and c.name in column_names:
                return self.on_conflict_update_string(
                    [c.name],
                    self.update_columns([c.name], column_names, arguments),
                )

        for c in self.columns:
            if c.primary_key and c.name in column_names:
                return self.on_conflict_update_string(
                    [c.name],
                    self.update_columns([c.name], column_names, arguments),
                )

        return ""

    def insert_command(self, column_names, arguments, conflict=None):
        substitution_string = self.substitution_string(arguments)

        if conflict == "update":
            upsert_string = self.create_upsert_string(column_names, arguments)
        else:
            upsert_string = ""

        return (
            "INSERT "
            + ("OR IGNORE " if conflict == "ignore" else "")
            + f"INTO {self.name} {self.column_string(column_names)} VALUES "
            + substitution_string
            + upsert_string
            + ";"
        )

    def column_names(self, **kwargs):
        return [c.name for c in self.columns if c.name in kwargs]

    def create_insert_args(self, **kwargs):
        column_names = self.column_names(**kwargs)
        arguments = [kwargs[name] for name in column_names]
        return (column_names, arguments)

    def insert(self, **kwargs):
        column_names, arguments = self.create_insert_args(**kwargs)
        return self.db.execute(
            self.insert_command(
                column_names, arguments, kwargs.get("conflict")
            ),
            arguments,
        )

    def insert_many(self, **kwargs):
        column_names, arguments = self.create_insert_args(**kwargs)
        self.db.execute_many(
            self.insert_command(
                column_names, arguments, kwargs.get("conflict")
            ),
            zip(*arguments),
        )

    def is_null_check(self, arg):
        if isinstance(arg, str) and "NULL" in arg:
            return True
        return False

    def equality_string(self, name, arg):
        if self.is_null_check(arg):
            return f"{name} IS {arg}"
        return f"{name} = ?"

    def create_where_string(self, column_names, arguments):
        where_string = (
            (
                " WHERE "
                + " AND ".join(
                    [
                        self.equality_string(name, arg)
                        for name, arg in zip(column_names, arguments)
                    ]
                )
            )
            if column_names
            else ""
        )
        arguments = [arg for arg in arguments if not self.is_null_check(arg)]
        return where_string, arguments

    def common_where_handling(self, **kwargs):
        return self.create_where_string(*self.create_insert_args(**kwargs))

    def select(self, column_string, **kwargs):
        where_string, arguments = self.common_where_handling(**kwargs)
        return self.db.execute(
            f"SELECT {column_string} FROM {self.name}{where_string};",
            arguments,
        )

    def select_where_in(self, field, values, column_string):
        return self.db.execute(
            f"SELECT {column_string} FROM {self.name} WHERE {field} IN ("
            + ("?, " * len(values))[:-2]
            + ");",
            values,
        )

    def delete(self, **kwargs):
        where_string, arguments = self.common_where_handling(**kwargs)
        self.db.execute(
            f"DELETE FROM {self.name}{where_string};",
            arguments,
        )

    def update(self, updates, where):
        where_string, arguments = self.common_where_handling(**where)

        update_args = []
        update_strings = []
        for k, v in updates.items():
            update_strings.append(f"{k} = ?")
            update_args.append(v)
        
        arguments = update_args + arguments

        self.db.execute(
            f"UPDATE {self.name} SET "
            + ", ".join(update_strings)
            + where_string
            + ";",
            arguments
        )

class KeyStoredObject:
    # Singletons like Sources or Targets which are referred in the database
    # by keys.

    def __init__(self, key):
        self.key: str = key


class StoredObject:
    # Objects which are stored in the database may subclass from this class.
    # If they do, they should have attributes for each of the relevant fields
    # in the database. id is provided by default and so only tables with an
    # id column are applicable.

    def __init__(self, table, db_id=None):
        self.table: str = table
        self._id: int = db_id

    @property
    def id(self):
        """Database ID of the StoredObject. Will store() to generate one."""

        if not self._id:
            self.store()
        return self._id

    def get_value(self, name):
        if name == "id":
            return self._id
        value = getattr(self, name, None)
        if isinstance(value, StoredObject):
            value = value.id
        elif isinstance(value, KeyStoredObject):
            value = value.key
        return value

    def store(self):
        """Store this object in the database."""

        kwargs = {}
        for column in database.tables[self.table].columns:
            kwargs[column.name] = self.get_value(column.name)

        if self._id:
            update(
                self.table,
                {
                    column.name: self.get_value(column.name)
                    for column in database.tables[self.table].columns
                    if column.name != "id"
                },
                { "id": self._id }
            )
        else:
            insert_id = upsert(self.table, **kwargs)
            if insert_id:
                self._id = insert_id
            elif not self._id:
                self._id = select_one_column(
                    self.table,
                    "id",
                    **{
                        column.name: self.get_value(column.name)
                        for column in database.tables[self.table].columns
                        if column.name != "id"
                    },
                )

    def delete_stored(self):
        """If this object has been stored in the db, delete the record."""
        if self._id:
            delete(self.table, id=self._id)


class DatabaseEvents(enum.Enum):
    CARD_LIST_UPDATE = 1


database = Database(
    [
        Table(
            "sources",
            [
                Column("short", "TEXT", primary_key=True),
                Column("name", "TEXT", not_null=True, unique=True),
            ],
        ),
        Table(
            "users",
            [
                Column("id", "INTEGER", primary_key=True),
                Column("name", "TEXT", not_null=True),
                Column("source", "TEXT", references="sources", not_null=True),
                Column("source_id", "TEXT"),
            ],
            ["UNIQUE(source, name)"],
        ),
        Table(
            "targets",
            [
                Column("short", "TEXT", primary_key=True),
                Column("name", "TEXT", not_null=True, unique=True),
            ],
        ),
        Table(
            "output_dirs",
            [
                Column("id", "INTEGER", primary_key=True),
                Column("path", "TEXT", not_null=True, unique=True),
            ],
        ),
        Table(
            "outputs",
            [
                Column("id", "INTEGER", primary_key=True),
                Column("target", "TEXT", references="targets", not_null=True),
                Column(
                    "output_dir",
                    "INTEGER",
                    references="output_dirs",
                    not_null=True,
                ),
                Column(
                    "profile", "INTEGER", references="profiles", not_null=True
                ),
                Column("include_maybe", "INTEGER")
            ],
            ["UNIQUE(target, output_dir, profile)"],
        ),
        Table(
            "profiles",
            [
                Column("id", "INTEGER", primary_key=True),
                Column("user", "INTEGER", references="users", not_null=True),
                Column("name", "TEXT", unique=True),
            ],
        ),
        Table(
            "cards",
            [
                Column("id", "INTEGER", primary_key=True),
                Column("name", "TEXT", index_on=True),
                Column("mtgo_id", "INTEGER", unique=True),
                Column("is_dfc", "INTEGER", not_null=True),
                Column("collector_number", "TEXT", not_null=True),
                Column("edition", "TEXT", not_null=True),
                Column("reprint", "INTEGER", not_null=True),
            ],
        ),
        Table(
            "decks",
            [
                Column("id", "INTEGER", primary_key=True),
                Column("deck_id", "TEXT", not_null=True, index_on=True),
                Column("source", "TEXT", not_null=True, references="sources"),
            ],
            ["UNIQUE(deck_id, source)"],
        ),
        Table(
            "deck_files",
            [
                Column("id", "INTEGER", primary_key=True),
                Column("deck", "INTEGER", references="decks", not_null=True),
                Column("file_name", "TEXT", not_null=True),
                Column(
                    "output",
                    "INTEGER",
                    references="outputs",
                    not_null=True,
                ),
                Column("updated", "INTEGER"),
            ],
            ["UNIQUE(file_name, output)"],
        ),
        Table(
            "database_events",
            [
                Column("id", "INTEGER", primary_key=True),
                Column("time", "INTEGER", not_null=True),
                Column("data", "TEXT"),
            ],
        ),
        Table(
            "string_values",
            [
                Column("key", "TEXT", primary_key=True),
                Column("value", "TEXT"),
            ],
        ),
    ],
)

insert = database.insert
insert_many = database.insert_many
insert_many_tuples = database.insert_many_tuples
upsert = database.upsert
upsert_many = database.upsert_many
select = database.select
select_one = database.select_one
select_one_column = database.select_one_column
select_ignore_none = database.select_ignore_none
delete = database.delete
update = database.update
execute = database.execute
commit = database.commit
close = database.close
enable_logging = database.enable_logging
disable_logging = database.disable_logging


def init():
    """Connect to the database, setting it up if necessary."""

    database_file = os.path.join(utils.DATA_DIR, DATABASE_FILE)
    initial_setup = not os.path.exists(database_file)
    utils.ensure_data_dir()
    database.init(database_file, initial_setup)

    disable_logging()
    from . import sources

    for source in sources.sourcelist:
        insert(
            "sources", conflict="ignore", short=source.SHORT, name=source.NAME
        )

    from . import targets

    for target in targets.targetlist:
        insert(
            "targets", conflict="ignore", short=target.SHORT, name=target.NAME
        )
    enable_logging()

    commit()



import logging
import time
import requests

from .. import utils
from ..modes import cli

from . import source


class Moxfield(source.Source):
    NAME = "Moxfield"
    SHORT = NAME[0]
    URL_BASE = "https://api.moxfield.com/"
    DECK_LIST_PAGE_SIZE = 100
    REQUEST_OK = 200
    USER_AGENT_KEY = "moxfield_user_agent"

    def __init__(self):
        super().__init__(Moxfield.NAME, Moxfield.SHORT)

        self._user_agent = ""
        self._last_request_time = 0
        self._logged_wait = False
        self._printed_user_agent_message = False

    def _request(self, url, *, params=None):
        delta = time.time() - self._last_request_time
        wait_time = round(1.0 - delta, 2)
        if wait_time > 0.0:
            if not self._logged_wait:
                logging.info(
                    "Waiting "
                    + str(wait_time)
                    + "s to obey Moxfield rate limit of 1 request per second."
                    + " Future waits will not be logged."
                )
                self._logged_wait = True
            time.sleep(wait_time)
        self._last_request_time = time.time()

        return requests.get(
            url, params=params, headers={"User-Agent": self._user_agent}
        )

    def parse_to_cards(self, board):
        cards = []
        for k in board:
            cards.append(
                (
                    board[k]["quantity"],
                    k,
                )
            )

        return cards

    def deck_to_generic_format(self, deck_id, deck):
        d = self.create_deck(deck_id, deck["name"], deck["description"])

        for board in ["mainboard", "sideboard", "maybeboard", "commanders"]:
            d.add_cards(self.parse_to_cards(deck.get(board, {})), board)

        return d

    def _get_deck(self, deck_id):
        return self.deck_to_generic_format(
            deck_id,
            self._request(f"{Moxfield.URL_BASE}v2/decks/all/{deck_id}").json(),
        )

    def deck_list_to_generic_format(self, decks):
        ret = []
        for deck in decks:
            ret.append(
                self.deck_update_from(
                    deck["publicId"],
                    utils.parse_iso_8601(deck["lastUpdatedAtUtc"]),
                )
            )
        return ret

    def _get_deck_list(self, username, allpages=True):
        decks = []
        i = 1
        while True:
            j = self._request(
                f"{Moxfield.URL_BASE}v2/users/{username}/decks",
                params={
                    "pageSize": Moxfield.DECK_LIST_PAGE_SIZE,
                    "pageNumber": i,
                },
            ).json()
            decks.extend(j["data"])
            i += 1
            if i > j["totalPages"] or not allpages:
                break

        return self.deck_list_to_generic_format(decks)

    def _verify_user(self, username):
        resp = self._request(f"{Moxfield.URL_BASE}v1/users/{username}")
        return resp.status_code == Moxfield.REQUEST_OK

    def ensure_setup(self, interactive, cache):
        if not self._user_agent:
            self._user_agent = cache.load_string_value(Moxfield.USER_AGENT_KEY)
        if not self._user_agent:
            if interactive:
                if not self._printed_user_agent_message:
                    print(USER_AGENT_MESSAGE)
                    self._printed_user_agent_message = True
                self._user_agent = cli.get_string("Moxfield user agent")
                resp = self._request(Moxfield.URL_BASE + "v1")
                if resp.status_code == 403:
                    prompt = (
                        "Failed to perform a request with user agent `"
                        + self._user_agent
                        + "`, try again?"
                    )
                    if cli.get_decision(prompt):
                        self.ensure_setup(True, cache)
                    else:
                        raise ValueError("Moxfield user agent not provided.")
                else:
                    cache.save_string_value(
                        Moxfield.USER_AGENT_KEY, self._user_agent
                    )
            else:
                raise ValueError(
                    "Missing Moxfield user agent and unable to retrieve it "
                    + "while non-interactive. Please run "
                    + utils.APP_NAME
                    + " interactively to provide your user agent."
                )


USER_AGENT_MESSAGE = """
The Moxfield API no longer allows arbritrary public access. As of version
0.1.0 architrice requires a user agent provided by Moxfield support to be able
to download decks from Moxfield. Such a user agent token can be obtained by
contacting Moxfield support by emailing support@moxfield.com with a request
similar to the below example:

Hi, I use a tool called Architrice (https://pypi.org/project/architrice/) to
sync my MTG decklists from Moxfield to my computer. This tool depends on the
Moxfield API and is blocked by Cloudflare. May I please have a user agent to
access the API so I can continue to use Moxfield in this way?

Once you have a user agent key (something like "architrice abcdef123456") you
can provide it for Architrice to use to download your decks.
"""



from architrice.targets.card_info import CARD_LIST_UPDATE_INTERVAL
import os
import xml.etree.cElementTree as et

from .. import utils

from . import target


# TODO Desktop/cockatrice_portable/data/decks
class Cockatrice(target.Target):
    NAME = "Cockatrice"
    SHORT = "C"
    DECK_DIRECTORY = (
        utils.expand_path(
            os.path.join(
                os.getenv("LOCALAPPDATA"), "Cockatrice/Cockatrice/decks"
            )
        )
        if os.name == "nt"
        else utils.expand_path("~/.local/share/Cockatrice/Cockatrice/decks")
    )
    DECK_FILE_EXTENSION = ".cod"
    SHORTCUT_NAME = "Cockatrice.lnk"
    EXECUTABLE_NAME = "cockatrice"
    SUPPORTS_RELNK = True

    def __init__(self):
        super().__init__(
            Cockatrice.NAME,
            Cockatrice.SHORT,
            Cockatrice.DECK_FILE_EXTENSION,
        )

    def suggest_directory(self):
        return Cockatrice.DECK_DIRECTORY

    def _save_deck(self, deck, path, include_maybe=False, card_info_map=None):
        self.deck_to_xml(deck, path, include_maybe, card_info_map)

    def deck_to_xml(self, deck, outfile, include_maybe, card_info_map=None):
        root = et.Element("cockatrice_deck", version="1")

        et.SubElement(root, "deckname").text = deck.name
        et.SubElement(root, "comments").text = deck.description

        main = et.SubElement(root, "zone", name="main")
        side = et.SubElement(root, "zone", name="side")

        for quantity, name in deck.get_main_deck():
            et.SubElement(
                main,
                "card",
                number=str(quantity),
                name=self.front_face_name(name, card_info_map),
            )
        for quantity, name in deck.get_sideboard(include_maybe=include_maybe):
            et.SubElement(
                side,
                "card",
                number=str(quantity),
                name=self.front_face_name(name, card_info_map),
            )

        et.ElementTree(root).write(
            outfile, xml_declaration=True, encoding="UTF-8"
        )



import re

import bs4
import requests

from .. import utils

from . import source


class Deckstats(source.Source):
    NAME = "Deckstats"
    SHORT = NAME[0]
    URL_BASE = "https://deckstats.net/"

    def __init__(self):
        super().__init__(Deckstats.NAME, Deckstats.SHORT)

    def format_deck_id(self, deck_id, owner_id):
        return f"{deck_id}&owner_id={owner_id}"

    def card_json_to_card(self, card):
        return (card["amount"], card["name"])

    def deck_to_generic_format(self, deck_id, deck):
        d = self.create_deck(deck_id, deck["name"], "")

        for section in deck.get("sections", []):
            for card in section.get("cards", []):
                if card.get("isCommander", False):
                    d.commanders.append(self.card_json_to_card(card))
                else:
                    d.main.append(self.card_json_to_card(card))

        for board in ["sideboard", "maybeboard"]:
            for card in deck.get(board, []):
                d.add_card(self.card_json_to_card(card), board)

        return d

    # API Reference:
    # https://deckstats.net/forum/index.php/topic,41323.msg112773.html#msg112773
    #
    # Note that deckstats IDs have the format "DECK_ID&owner_id=OWNER_ID"
    # because this app assumes decks can be retrieved from a single ID, while
    # deckstats requires owner ID as well.
    #
    # It is for this reason that params is not used in requests.get, as it would
    # escape ampersands in the id.
    def _get_deck(self, deck_id):
        return self.deck_to_generic_format(
            deck_id,
            requests.get(
                Deckstats.URL_BASE
                + "api.php/?action=get_deck&id_type=saved&response_type=json"
                f"&id={deck_id}",
            ).json(),
        )

    def get_user_id(self, username):
        html = requests.get(
            f"{Deckstats.URL_BASE}members/search/?search_name={username}"
        ).content.decode()
        soup = bs4.BeautifulSoup(html, "html.parser")
        try:
            href = soup.select_one("a.member_name").get("href")
            return re.sub(
                r"^https://deckstats\.net/decks/(\d+).*$", r"\1", href
            )
        except AttributeError:
            return None

    def _get_deck_list(self, username):
        user_id = self.get_user_id(username)

        decks = []
        i = 1
        while True:
            data = requests.get(
                Deckstats.URL_BASE + "api.php",
                params={
                    "decks_page": i,
                    "owner_id": user_id,
                    "action": "user_folder_get",
                    "result_type": "folder;decks;parent_tree;subfolders",
                },
            ).json()

            folder = data.get("folder")
            if folder:
                for deck in folder.get("decks", []):
                    decks.append(
                        self.deck_update_from(
                            self.format_deck_id(str(deck["saved_id"]), user_id),
                            utils.timestamp_to_utc(deck["updated"]),
                        )
                    )

                if (
                    folder["decks_current_page"] * folder["decks_per_page"]
                    < folder["decks_total"]
                ):
                    i += 1
                else:
                    return decks
            else:
                return []

    def _verify_user(self, username):
        return bool(self.get_user_id(username))



import json
import logging
import os

from . import cli
from . import common_json
from . import mode


class ImportJson(mode.Mode):
    def __init__(self):
        super().__init__(
            "J",
            "import-json",
            "import JSON file (-p) as profile list",
            ["path"],
        )

    def resolve_missing_arg(self, cache, arg, args):
        if arg != "path":
            raise ValueError(f"Can't resolve missing argument: {arg}.")

        return cli.get_path("JSON file to import")

    def action(self, cache, args):
        if os.path.isfile(args.path):
            with open(args.path, "r") as f:
                try:
                    data = json.load(f)
                except json.decoder.JSONDecodeError:
                    logging.error(f"Failed to parse JSON file {args.path}.")
                    return

            if isinstance(data, list):
                # should be a list of profiles
                for i, profile in enumerate(data, 1):
                    identifier = f"Profile {i}"
                    if not isinstance(profile, dict):
                        logging.error(
                            f"Profiles should be key-value. {identifier}"
                            " is not. Aborting."
                        )
                        break
                    elif not common_json.verify_profile_json(profile):
                        identifier = profile.get("name") or identifier
                        logging.error(f"{identifier} is invalid. Aborting.")
                        break
                else:
                    for i, profile in enumerate(data, 1):
                        identifier = profile.get("name") or f"Profile {i}"
                        common_json.import_profile_json(cache, profile)
                        logging.info(f"{identifier} added.")
            elif common_json.verify_profile_json(data):
                # accept a single profile
                common_json.import_profile_json(cache, data)
            else:
                logging.error(
                    "File must contain a single profile or a list of profiles,"
                    " in JSON format."
                )



import abc
import os

from .. import database
from .. import utils

from . import card_info


class Target(database.KeyStoredObject, abc.ABC):
    SUPPORTS_RELNK = False

    def __init__(self, name, short, file_extension):
        database.KeyStoredObject.__init__(self, short)

        self.name = name
        self.short = short
        self.file_extension = file_extension
        self.mtgo_id_required = False

    def suggest_directory(self):
        if os.name == "nt":
            return utils.expand_path(
                os.path.join(os.getenv("USERPROFILE"), "Documents", "Decks")
            )
        else:
            return utils.expand_path(os.path.join("~", "Decks"))

    @abc.abstractmethod
    def _save_deck(self, deck, path, include_maybe=False, card_info_map=None):
        """Writes deck to path in format using card_info_map."""

    def front_face_name(self, name, card_info_map=None):
        if card_info_map and name in card_info_map:
            card = card_info_map[name]
            if card.is_dfc:
                return name.partition("//")[0].strip()
        return name

    def save_deck(self, deck, path, include_maybe=False, card_info_map=None):
        if card_info_map is None:
            card_info_map = card_info.map_from_deck(deck)
        self._save_deck(deck, path, include_maybe, card_info_map)

    def save_decks(self, deck_tuples, include_maybe=False, card_info_map=None):
        if card_info_map is None:
            card_info_map = card_info.map_from_decks(
                [d for d, _ in deck_tuples],
                mtgo_id_required=self.mtgo_id_required,
            )

        for deck, path in deck_tuples:
            self.save_deck(deck, path, include_maybe, card_info_map)

    def create_file_name(self, deck_name):
        return utils.create_file_name(deck_name) + self.file_extension



import setuptools

with open("architrice/__init__.py", "r") as f:
    for line in f:
        if "__version__" in line:
            version = line.split("=")[1].strip().replace('"', "")

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="architrice",
    version=version,
    url="https://github.com/OwenFeik/architrice.git",
    author="Owen Feik",
    author_email="owen.h.feik@gmail.com",
    description="Utility to sync MtG decklists with online sources.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    download_url="https://github.com/OwenFeik/architrice/archive/refs/tags/0.1.0.tar.gz",
    install_requires=["requests", "bs4"],
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Development Status :: 3 - Alpha",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
    ],
    python_requires=">=3.7",
)



import os
import tempfile
import unittest

import architrice

from . import mockapi


class TestIntegration(unittest.TestCase):
    TEST_USER_NAME = "Test"
    TEST_DECK_NAME = "Test Deck"
    TEST_DATA_DIR = os.path.join(os.path.dirname(__file__), "data")

    @classmethod
    def setUpClass(cls):
        cls.directory = tempfile.mkdtemp()
        print("Test output: " + cls.directory)

        architrice.utils._DATA_DIR = architrice.utils.DATA_DIR
        architrice.utils.DATA_DIR = os.path.join(cls.directory, "architrice")
        architrice.database.init()
        mockapi.mock()

    @classmethod
    def tearDown(cls):
        architrice.utils.DATA_DIR = architrice.utils._DATA_DIR
        architrice.database.close()
        mockapi.stop()

    def verifyOutput(self, profile):
        for output in profile.outputs:
            file_name = output.target.create_file_name(
                TestIntegration.TEST_DECK_NAME
            )
            created_file = os.path.join(
                output.output_dir.path,
                file_name,
            )

            failure_info = (
                f"for source {profile.source.short} and "
                f"target {output.target.short}."
            )

            self.assertTrue(
                os.path.exists(created_file),
                f"Deck file not created {failure_info}",
            )
            with open(created_file) as f:
                created_text = f.read()

            with open(
                os.path.join(
                    TestIntegration.TEST_DATA_DIR,
                    output.target.short,
                    file_name,
                )
            ) as f:
                correct_text = f.read()

            self.assertEqual(
                created_text,
                correct_text,
                f"Bad deck file {failure_info}: {created_file}",
            )

    def testIntegration(self):
        # Tests by creating a profile for each source, with an output for each
        # target and downloading the test deck for each source-target
        # combination, verifying that the decks were downloaded.

        cache = architrice.caching.Cache.load()

        deck_dir = os.path.join(TestIntegration.directory, "out")

        for source in architrice.sources.get_all():
            self.assertTrue(
                source.verify_user(TestIntegration.TEST_USER_NAME),
                f"User verification failed for source {source.short}",
            )
            profile = cache.build_profile(
                source, TestIntegration.TEST_USER_NAME
            )

            for target in architrice.targets.get_all():
                cache.build_output(
                    profile,
                    target,
                    os.path.join(deck_dir, source.short, target.short),
                    True,
                )

            profile.download_all()
            self.verifyOutput(profile)


if __name__ == "__main__":
    unittest.main()



from .archidekt import Archidekt
from .deckstats import Deckstats
from .moxfield import Moxfield
from .tappedout import TappedOut

sourcelist = [Archidekt, Deckstats, Moxfield, TappedOut]

# singleton cache; sources have no data so only a single instance is required
_sources = {}


def get(name, error_on_fail=False):
    if name:
        name = name.lower()  # case insensitivity
        for source in sourcelist:
            if source.NAME.lower() == name or source.SHORT.lower() == name:
                if source.SHORT not in _sources:
                    _sources[source.SHORT] = source()
                return _sources[source.SHORT]

    if error_on_fail:
        raise ValueError(f'No source matching "{name}" exists.')
    return None


def get_all():
    return list(map(lambda s: get(s.NAME), sourcelist))



import os

from . import target


class Generic(target.Target):
    NAME = "Generic"
    SHORT = "G"
    DECK_FILE_EXTENSION = ".txt"

    def __init__(self):
        super().__init__(
            Generic.NAME, Generic.SHORT, Generic.DECK_FILE_EXTENSION
        )

    def canonical_name(self, name, card_info_map):
        return card_info_map[name].name

    def _save_deck(self, deck, path, include_maybe, card_info_map):

        deck_string = ""
        for quantity, name in deck.get_main_deck():
            deck_string += (
                f"{quantity} {self.canonical_name(name, card_info_map)}\n"
            )
        deck_string += "\n"
        for quantity, name in deck.get_sideboard(include_maybe=include_maybe):
            deck_string += (
                f"{quantity} {self.canonical_name(name, card_info_map)}\n"
            )

        with open(path, "w") as f:
            f.write(deck_string)



import datetime
import itertools
import logging
import os
import re
import sys


APP_NAME = "architrice"
DEBUG = False


def get_data_dir():
    if DEBUG:
        return os.path.join(os.path.dirname(__file__), "data")

    if os.name == "nt":
        return os.path.join(os.getenv("LOCALAPPDATA"), "architrice")
    else:
        DATA_HOME_ENV_VAR = "XDG_DATA_HOME"
        DATA_HOME_FALLBACK = "~/.local/share"

        return os.path.join(
            os.path.expanduser(
                os.getenv(DATA_HOME_ENV_VAR) or DATA_HOME_FALLBACK
            ),
            "architrice",
        )


DATA_DIR = get_data_dir()
LOG_FILE = os.path.join(DATA_DIR, "architrice.log")


def ensure_data_dir():
    if not os.path.isdir(DATA_DIR):
        os.mkdir(DATA_DIR)


def set_up_logger(verbosity=1):
    ensure_data_dir()
    handlers = [logging.FileHandler(LOG_FILE)]

    if verbosity != 0:
        stdout_handler = logging.StreamHandler(sys.stdout)
        stdout_handler.setFormatter(
            logging.Formatter("%(levelname)s: %(message)s")
        )
        stdout_handler.setLevel(
            logging.INFO if verbosity == 1 else logging.DEBUG
        )
        handlers.append(stdout_handler)

    logging.basicConfig(level=logging.DEBUG, handlers=handlers)


def create_file_name(deck_name):
    return re.sub("[^a-z0-9_ ]+", "", deck_name.lower()).replace(" ", "_")


def time_now():
    return int(datetime.datetime.utcnow().timestamp())


def timestamp_to_utc(timestamp):
    return int(datetime.datetime.utcfromtimestamp(timestamp).timestamp())


def parse_iso_8601(time_string):
    return int(
        datetime.datetime.strptime(
            time_string, "%Y-%m-%dT%H:%M:%S.%fZ"
        ).timestamp()
    )


def expand_path(path):
    if path is None:
        return None
    return os.path.abspath(os.path.expanduser(path))


def check_dir(path):
    if not path:
        return False
    if os.path.isfile(path):
        return False
    if not os.path.isdir(path):
        os.makedirs(path)
        logging.info(f"Created output directory {path}.")
    return True


# Chunk an iterable into n size chunks
# source: https://docs.python.org/3/library/itertools.html#itertools-recipes
def group_iterable(iterable, n, fillvalue=None):
    args = [iter(iterable)] * n
    return itertools.zip_longest(*args, fillvalue=fillvalue)
